from __future__ import annotations

import contextlib
import typing
from abc import ABCMeta, abstractmethod

from libqtile import hook
from libqtile.command.base import CommandObject, expose_command
from libqtile.config import Screen

if typing.TYPE_CHECKING:
    from typing import Any

    from libqtile import config
    from libqtile.backend.base import Internal
    from libqtile.backend.base.idle_inhibit import IdleInhibitorManager
    from libqtile.backend.base.idle_notify import IdleNotifier
    from libqtile.command.base import ItemT
    from libqtile.core.manager import Qtile
    from libqtile.group import _Group


class Core(CommandObject, metaclass=ABCMeta):
    painter: Any
    supports_restarting: bool = True
    qtile: Qtile
    idle_inhibitor_manager: IdleInhibitorManager[Any]
    idle_notifier: IdleNotifier

    @property
    @abstractmethod
    def name(self) -> str:
        """The name of the backend"""

    def _items(self, name: str) -> ItemT:
        return None

    def _select(self, name, sel):
        return None

    @abstractmethod
    def finalize(self):
        """Destructor/Clean up resources"""

    @property
    @abstractmethod
    def display_name(self) -> str:
        pass

    @abstractmethod
    def setup_listener(self) -> None:
        """Setup a listener for the given qtile instance"""

    @abstractmethod
    def remove_listener(self) -> None:
        """Setup a listener for the given qtile instance"""

    def update_desktops(self, groups: list[_Group], index: int) -> None:
        """Set the current desktops of the window manager"""

    @abstractmethod
    def get_output_info(self) -> list[config.Output]:
        """Get the output information"""

    @abstractmethod
    def grab_key(self, key: config.Key | config.KeyChord) -> tuple[int, int]:
        """Configure the backend to grab the key event"""

    @abstractmethod
    def ungrab_key(self, key: config.Key | config.KeyChord) -> tuple[int, int]:
        """Release the given key event"""

    @abstractmethod
    def ungrab_keys(self) -> None:
        """Release the grabbed key events"""

    @abstractmethod
    def grab_button(self, mouse: config.Mouse) -> int:
        """Configure the backend to grab the mouse event"""

    def ungrab_buttons(self) -> None:
        """Release the grabbed button events"""

    def grab_pointer(self) -> None:
        """Configure the backend to grab mouse events"""

    def ungrab_pointer(self) -> None:
        """Release grabbed pointer events"""

    def on_config_load(self, initial: bool) -> None:
        """
        Respond to config loading. `initial` will be `True` if Qtile just started.
        """

    def warp_pointer(self, x: int, y: int) -> None:
        """Warp the pointer to the given coordinates relative."""

    @contextlib.contextmanager
    def masked(self):
        """A context manager to suppress window events while operating on many windows."""
        yield

    def create_internal(self, x: int, y: int, width: int, height: int) -> Internal:
        """Create an internal window controlled by Qtile."""
        raise NotImplementedError  # Only error when called, not when instantiating class

    def flush(self) -> None:
        """If needed, flush the backend's event queue."""

    def simulate_keypress(self, modifiers: list[str], key: str) -> None:
        """Simulate a keypress with given modifiers"""

    def keysym_from_name(self, name: str) -> int:
        """Get the keysym for a key from its name"""
        raise NotImplementedError

    def get_mouse_position(self) -> tuple[int, int]:
        """Get mouse coordinates."""
        raise NotImplementedError

    @expose_command()
    def info(self) -> dict[str, Any]:
        """Get basic information about the running backend."""
        return {"backend": self.name, "display_name": self.display_name}

    def check_screen_fullscreen_background(self, screen: Screen | None = None) -> None:
        """Toggles fullscreen background if any window on the screen is fullscreen."""
        # Wayland only

    def update_backend_log_level(self) -> None:
        """Update the backend log level based on Qtile's log level."""
        # Wayland only

    @property
    def inhibited(self) -> bool:
        if not hasattr(self, "_inhibited"):
            self._inhibited = False

        return self._inhibited

    @inhibited.setter
    def inhibited(self, value: bool):
        if value != self.inhibited:
            self._inhibited = value
            hook.fire("idle_inhibitor_change", value)

    @expose_command()
    def set_idle_inhibitor(self) -> None:
        """Create a global idle inhibitor."""
        self.idle_inhibitor_manager.add_global_inhibitor()

    @expose_command()
    def remove_idle_inhibitor(self) -> None:
        """Remove global idle inhibitor."""
        self.idle_inhibitor_manager.remove_global_inhibitor()

    @expose_command()
    def get_idle_inhibitors(self, active_only: bool = False) -> list[str]:
        """Return list of inhibitors."""
        return [
            f"{inhibitor!r}"
            for inhibitor in self.idle_inhibitor_manager.inhibitors
            if not active_only or (active_only and inhibitor.check())
        ]



"""
Slice layout. Serves as example of delegating layouts (or sublayouts)
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from libqtile.backend.base import Window
from libqtile.command.base import expose_command
from libqtile.config import ScreenRect
from libqtile.layout.base import Layout
from libqtile.layout.max import Max

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any, Self

    from libqtile.group import _Group


class Single(Layout):
    """Layout with single window

    Just like Max but asserts that window is the one
    """

    def __init__(self):
        Layout.__init__(self)
        self.window = None
        self.focused = False

    def add_client(self, window):
        assert self.window is None
        self.window = window

    def remove(self, window: Window) -> None:
        if self.window is not window:
            raise ValueError("Cannot remove, window not managed by layout")
        self.window = None

    def configure(self, window, screen_rect):
        if window is self.window:
            window.place(
                screen_rect.x,
                screen_rect.y,
                screen_rect.width,
                screen_rect.height,
                0,
                None,
            )
            window.unhide()
        else:
            window.hide()

    def empty(self):
        """Is the layout empty

        Returns True if the layout empty (and is willing to accept windows)
        """
        return self.window is None

    def focus_first(self) -> Window | None:
        self.focused = True
        return self.window

    def focus_last(self) -> Window | None:
        self.focused = True
        return self.window

    def focus_next(self, window: Window) -> Window | None:
        if self.focused:
            self.focused = False
            return None
        return self.window

    def focus_previous(self, window: Window) -> Window | None:
        if self.focused:
            self.focused = False
            return None
        return self.window

    def next(self) -> None:
        pass

    def previous(self) -> None:
        pass

    def get_windows(self):
        return self.window

    def info(self) -> dict[str, Any]:
        d = Layout.info(self)
        d["window"] = self.window.name if self.window else ""
        return d


class Slice(Layout):
    """Slice layout

    This layout cuts piece of screen_rect and places a single window on that
    piece, and delegates other window placement to other layout
    """

    defaults = [
        ("width", 256, "Slice width."),
        ("side", "left", "Position of the slice (left, right, top, bottom)."),
        ("match", None, "Match-object describing which window(s) to move to the slice."),
        ("fallback", Max(), "Layout to be used for the non-slice area."),
    ]

    fallback: Layout

    def __init__(self, **config):
        self.layouts = {}
        Layout.__init__(self, **config)
        self.add_defaults(Slice.defaults)
        self._slice = Single()

    def clone(self, group: _Group) -> Self:
        res = Layout.clone(self, group)
        res._slice = self._slice.clone(group)
        res.fallback = self.fallback.clone(group)
        return res

    def delegate_layout(self, windows, mapping):
        """Delegates layouting actual windows

        Parameters
        ===========
        windows:
            windows to layout
        mapping:
            mapping from layout to ScreenRect for each layout
        """
        grouped = {}
        for w in windows:
            lay = self.layouts[w]
            if lay in grouped:
                grouped[lay].append(w)
            else:
                grouped[lay] = [w]
        for lay, wins in grouped.items():
            lay.layout(wins, mapping[lay])

    def layout(self, windows: Sequence[Window], screen_rect: ScreenRect) -> None:
        win, sub = self._get_screen_rects(screen_rect)
        self.delegate_layout(
            windows,
            {
                self._slice: win,
                self.fallback: sub,
            },
        )

    def show(self, screen_rect: ScreenRect) -> None:
        win, sub = self._get_screen_rects(screen_rect)
        self._slice.show(win)
        self.fallback.show(sub)

    def configure(self, win, screen_rect):
        raise NotImplementedError("Should not be called")

    def _get_layouts(self):
        return (self._slice, self.fallback)

    def _get_active_layout(self):
        return self.fallback  # always

    def _get_screen_rects(self, screen):
        if self.side == "left":
            win, sub = screen.hsplit(self.width)
        elif self.side == "right":
            sub, win = screen.hsplit(screen.width - self.width)
        elif self.side == "top":
            win, sub = screen.vsplit(self.width)
        elif self.side == "bottom":
            sub, win = screen.vsplit(screen.height - self.width)
        else:
            raise NotImplementedError(self.side)
        return (win, sub)

    def add_client(self, win):
        if self._slice.empty() and self.match and self.match.compare(win):
            self._slice.add_client(win)
            self.layouts[win] = self._slice
        else:
            self.fallback.add_client(win)
            self.layouts[win] = self.fallback

    def remove(self, win: Window) -> Window:
        lay = self.layouts.pop(win)
        focus = lay.remove(win)
        if not focus:
            layouts = self._get_layouts()
            idx = layouts.index(lay)
            while idx < len(layouts) - 1 and not focus:
                idx += 1
                focus = layouts[idx].focus_first()
        return focus

    def hide(self) -> None:
        for lay in self._get_layouts():
            lay.hide()

    def focus(self, win):
        self.layouts[win].focus(win)

    def blur(self) -> None:
        for lay in self._get_layouts():
            lay.blur()

    def focus_first(self) -> Window | None:
        layouts = self._get_layouts()
        for lay in layouts:
            win = lay.focus_first()
            if win:
                return win
        return None

    def focus_last(self) -> None:
        layouts = self._get_layouts()
        for lay in reversed(layouts):
            win = lay.focus_last()
            if win:
                return win
        return None

    def focus_next(self, win: Window) -> Window | None:
        layouts = self._get_layouts()
        cur = self.layouts[win]
        focus = cur.focus_next(win)
        if not focus:
            idx = layouts.index(cur)
            while idx < len(layouts) - 1 and not focus:
                idx += 1
                focus = layouts[idx].focus_first()
        return focus

    def focus_previous(self, win: Window) -> Window | None:
        layouts = self._get_layouts()
        cur = self.layouts[win]
        focus = cur.focus_previous(win)
        if not focus:
            idx = layouts.index(cur)
            while idx > 0 and not focus:
                idx -= 1
                focus = layouts[idx].focus_last()
        return focus

    def __getattr__(self, name):
        """Delegate unimplemented command calls to active layout.

        For exposed commands that don't exist on the Slice class, this looks
        for an implementation on the active layout.
        """
        if "fallback" in self.__dict__:
            cmd = self.command(name)
            if cmd:
                return cmd
        return super().__getattr__(name)

    @expose_command()
    def next(self) -> None:
        self.fallback.next()

    @expose_command()
    def previous(self) -> None:
        self.fallback.previous()

    @expose_command()
    def commands(self):
        cmds = self._get_active_layout().commands()
        cmds.extend(cmd for cmd in Layout.commands(self) if cmd not in cmds)
        return cmds

    def get_windows(self):
        clients = list()
        for layout in self._get_layouts():
            if layout.get_windows() is not None:
                clients.extend(layout.get_windows())
        return clients

    def command(self, name: str):
        if name in self._commands:
            return self._commands.get(name)

        elif name in self._get_active_layout()._commands:
            return getattr(self._get_active_layout(), name)

    @expose_command()
    def move_to_slice(self):
        """Moves the current window to the slice."""
        win = self.group.current_window
        old_slice = self._slice.window
        if old_slice:
            self._slice.remove(old_slice)
            self.fallback.add_client(old_slice)
            self.layouts[old_slice] = self.fallback
        self.fallback.remove(win)
        self._slice.add_client(win)
        self.layouts[win] = self._slice
        self.group.layout_all()

    @expose_command()
    def info(self) -> dict[str, Any]:
        d = Layout.info(self)
        for layout in self._get_layouts():
            d[layout.name] = layout.info()
        return d



"""
A minimal EWMH-aware OO layer over xcffib. This is NOT intended to be
complete - it only implements the subset of functionalty needed by qtile.
"""

from __future__ import annotations

import contextlib
import functools
import operator
import struct
from itertools import chain, repeat

import cairocffi
import cairocffi.pixbuf
import cairocffi.xcb
import xcffib
import xcffib.randr
import xcffib.screensaver
import xcffib.xinerama
import xcffib.xproto
from xcffib.xfixes import SelectionEventMask
from xcffib.xproto import CW, EventMask, WindowClass

from libqtile.backend.x11 import window
from libqtile.backend.x11.xcursors import Cursors
from libqtile.backend.x11.xkeysyms import keysyms
from libqtile.config import Output, ScreenRect
from libqtile.log_utils import logger
from libqtile.utils import QtileError, hex, rgb


class XCBQError(QtileError):
    pass


# Keyboard modifiers bitmask values from X Protocol
ModMasks = {
    "shift": 1 << 0,
    "lock": 1 << 1,
    "control": 1 << 2,
    "mod1": 1 << 3,
    "mod2": 1 << 4,
    "mod3": 1 << 5,
    "mod4": 1 << 6,
    "mod5": 1 << 7,
}

AllButtonsMask = 0b11111 << 8
ButtonMotionMask = 1 << 13
ButtonReleaseMask = 1 << 3
PointerMotionHintMask = 1 << 7

NormalHintsFlags = {
    "USPosition": 1,  # User-specified x, y
    "USSize": 2,  # User-specified width, height
    "PPosition": 4,  # Program-specified position
    "PSize": 8,  # Program-specified size
    "PMinSize": 16,  # Program-specified minimum size
    "PMaxSize": 32,  # Program-specified maximum size
    "PResizeInc": 64,  # Program-specified resize increments
    "PAspect": 128,  # Program-specified min and max aspect ratios
    "PBaseSize": 256,  # Program-specified base size
    "PWinGravity": 512,  # Program-specified window gravity
}

HintsFlags = {
    "InputHint": 1,  # input
    "StateHint": 2,  # initial_state
    "IconPixmapHint": 4,  # icon_pixmap
    "IconWindowHint": 8,  # icon_window
    "IconPositionHint": 16,  # icon_x & icon_y
    "IconMaskHint": 32,  # icon_mask
    "WindowGroupHint": 64,  # window_group
    "MessageHint": 128,  # (this bit is obsolete)
    "UrgencyHint": 256,  # urgency
}

# http://standards.freedesktop.org/wm-spec/latest/ar01s05.html#idm139870830002400
WindowTypes = {
    "_NET_WM_WINDOW_TYPE_DESKTOP": "desktop",
    "_NET_WM_WINDOW_TYPE_DOCK": "dock",
    "_NET_WM_WINDOW_TYPE_TOOLBAR": "toolbar",
    "_NET_WM_WINDOW_TYPE_MENU": "menu",
    "_NET_WM_WINDOW_TYPE_UTILITY": "utility",
    "_NET_WM_WINDOW_TYPE_SPLASH": "splash",
    "_NET_WM_WINDOW_TYPE_DIALOG": "dialog",
    "_NET_WM_WINDOW_TYPE_DROPDOWN_MENU": "dropdown",
    "_NET_WM_WINDOW_TYPE_POPUP_MENU": "menu",
    "_NET_WM_WINDOW_TYPE_TOOLTIP": "tooltip",
    "_NET_WM_WINDOW_TYPE_NOTIFICATION": "notification",
    "_NET_WM_WINDOW_TYPE_COMBO": "combo",
    "_NET_WM_WINDOW_TYPE_DND": "dnd",
    "_NET_WM_WINDOW_TYPE_NORMAL": "normal",
}

# http://standards.freedesktop.org/wm-spec/latest/ar01s05.html#idm139870829988448
net_wm_states = (
    "_NET_WM_STATE_MODAL",
    "_NET_WM_STATE_STICKY",
    "_NET_WM_STATE_MAXIMIZED_VERT",
    "_NET_WM_STATE_MAXIMIZED_HORZ",
    "_NET_WM_STATE_SHADED",
    "_NET_WM_STATE_SKIP_TASKBAR",
    "_NET_WM_STATE_SKIP_PAGER",
    "_NET_WM_STATE_HIDDEN",
    "_NET_WM_STATE_FULLSCREEN",
    "_NET_WM_STATE_ABOVE",
    "_NET_WM_STATE_BELOW",
    "_NET_WM_STATE_DEMANDS_ATTENTION",
    "_NET_WM_STATE_FOCUSED",
)

WindowStates = {
    None: "normal",
    "_NET_WM_STATE_FULLSCREEN": "fullscreen",
    "_NET_WM_STATE_DEMANDS_ATTENTION": "urgent",
}

# Maps property names to types and formats.
PropertyMap = {
    # ewmh properties
    "_NET_DESKTOP_GEOMETRY": ("CARDINAL", 32),
    "_NET_SUPPORTED": ("ATOM", 32),
    "_NET_SUPPORTING_WM_CHECK": ("WINDOW", 32),
    "_NET_WM_NAME": ("UTF8_STRING", 8),
    "_NET_WM_PID": ("CARDINAL", 32),
    "_NET_CLIENT_LIST": ("WINDOW", 32),
    "_NET_CLIENT_LIST_STACKING": ("WINDOW", 32),
    "_NET_NUMBER_OF_DESKTOPS": ("CARDINAL", 32),
    "_NET_CURRENT_DESKTOP": ("CARDINAL", 32),
    "_NET_DESKTOP_NAMES": ("UTF8_STRING", 8),
    "_NET_DESKTOP_VIEWPORT": ("CARDINAL", 32),
    "_NET_WORKAREA": ("CARDINAL", 32),
    "_NET_ACTIVE_WINDOW": ("WINDOW", 32),
    "_NET_WM_DESKTOP": ("CARDINAL", 32),
    "_NET_WM_STRUT": ("CARDINAL", 32),
    "_NET_WM_STRUT_PARTIAL": ("CARDINAL", 32),
    "_NET_WM_WINDOW_OPACITY": ("CARDINAL", 32),
    "_NET_WM_WINDOW_TYPE": ("ATOM", 32),
    "_NET_FRAME_EXTENTS": ("CARDINAL", 32),
    # Net State
    "_NET_WM_STATE": ("ATOM", 32),
    # Xembed
    "_XEMBED_INFO": ("_XEMBED_INFO", 32),
    # ICCCM
    "WM_STATE": ("WM_STATE", 32),
    # Qtile-specific properties
    "QTILE_INTERNAL": ("CARDINAL", 32),
}
for _name in net_wm_states:
    PropertyMap[_name] = ("ATOM", 32)

# TODO add everything required here:
# http://standards.freedesktop.org/wm-spec/latest/ar01s03.html
SUPPORTED_ATOMS = [
    # From http://standards.freedesktop.org/wm-spec/latest/ar01s03.html
    "_NET_SUPPORTED",
    "_NET_CLIENT_LIST",
    "_NET_CLIENT_LIST_STACKING",
    "_NET_CURRENT_DESKTOP",
    "_NET_DESKTOP_VIEWPORT",
    "_NET_ACTIVE_WINDOW",
    "_NET_SUPPORTING_WM_CHECK",
    # From http://standards.freedesktop.org/wm-spec/latest/ar01s05.html
    "_NET_WM_NAME",
    "_NET_WM_VISIBLE_NAME",
    "_NET_WM_ICON_NAME",
    "_NET_WM_DESKTOP",
    "_NET_WM_WINDOW_TYPE",
    "_NET_WM_STATE",
    "_NET_WM_STRUT_PARTIAL",
    "_NET_WM_PID",
]
SUPPORTED_ATOMS.extend(WindowTypes.keys())
SUPPORTED_ATOMS.extend(net_wm_states)

XCB_CONN_ERRORS = {
    1: "XCB_CONN_ERROR",
    2: "XCB_CONN_CLOSED_EXT_NOTSUPPORTED",
    3: "XCB_CONN_CLOSED_MEM_INSUFFICIENT",
    4: "XCB_CONN_CLOSED_REQ_LEN_EXCEED",
    5: "XCB_CONN_CLOSED_PARSE_ERR",
    6: "XCB_CONN_CLOSED_INVALID_SCREEN",
    7: "XCB_CONN_CLOSED_FDPASSING_FAILED",
}

# Some opcodes from xproto.h, used for faking input.
XCB_KEY_PRESS = 2
XCB_KEY_RELEASE = 3
XCB_BUTTON_PRESS = 4
XCB_BUTTON_RELEASE = 5
XCB_MOTION_NOTIFY = 6


def parse_edid(raw):
    # https://en.wikipedia.org/wiki/Extended_Display_Identification_Data#EDID_1.4_data_format
    if sum(map(int, raw)) % 256 != 0:
        raise Exception(f"bad EDID data checksum {raw}")

    # Extract manufacturer ID from bytes 8-9 (3 letters encoded in 5 bits each)
    mfg_bytes = struct.unpack(">H", raw[8:10])[0]
    make = "".join(
        [
            chr(((mfg_bytes >> 10) & 0x1F) + ord("A") - 1),
            chr(((mfg_bytes >> 5) & 0x1F) + ord("A") - 1),
            chr((mfg_bytes & 0x1F) + ord("A") - 1),
        ]
    )

    # we only care about the serial number/monitor model name, which can occur
    # in any of the four possible timing locations, each of which are 18
    # bytes. We discard everything else.
    timings = struct.unpack("<54x18s18s18s18s2x", raw[:128])
    serial = None
    model = None
    for t in timings:
        # skip timing info. we could decode this, but we can get it via
        # randr's preferred mode encoding if we eventually care.
        if t[0:2] != b"\x00\x00":
            continue
        # https://en.wikipedia.org/wiki/Extended_Display_Identification_Data#Monitor_Descriptors
        content = t[5:].decode("cp437").strip()
        if t[3] == 0xFF:
            serial = content

        if t[3] == 0xFC:
            model = content

        # some monitors export *both* the serial and the model as 0xfe, which
        # is annoying. if we see this, first set the model, then set the
        # serial and hope for the best. if only they had used, you know, one
        # of the masks that indicated serial or model!
        if t[3] == 0xFE:
            content = t[5:].decode("cp437").strip()
            if model is None:
                model = content
            else:
                serial = content

    if serial == "000000000000":
        serial = None

    return (make, model, serial)


class MaskMap:
    """
    A general utility class that encapsulates the way the bitmask/listofvalue idiom
    works in X protocol. It understands a special attribute _maskvalue on
    objects, which will be used instead of the object value if present.
    This lets us pass in a Font object, rather than Font.fid, for example.
    """

    def __init__(self, obj):
        self.mmap = []
        for i in dir(obj):
            if not i.startswith("_"):
                self.mmap.append((getattr(obj, i), i.lower()))
        self.mmap.sort()

    def __call__(self, **kwargs):
        """
        kwargs: keys should be in the mmap name set

        Returns a (mask, values) tuple.
        """
        mask = 0
        values = []
        for m, s in self.mmap:
            if s in kwargs:
                val = kwargs.get(s)
                if val is not None:
                    mask |= m
                    values.append(getattr(val, "_maskvalue", val))
                del kwargs[s]
        if kwargs:
            raise ValueError(f"Unknown mask names: {list(kwargs.keys())}")
        return mask, values


ConfigureMasks = MaskMap(xcffib.xproto.ConfigWindow)
AttributeMasks = MaskMap(CW)


class AtomCache:
    def __init__(self, conn):
        self.conn = conn
        self.atoms = {}
        self.reverse = {}

        # We can change the pre-loads not to wait for a return
        for name in WindowTypes.keys():
            self.insert(name=name)

        for i in dir(xcffib.xproto.Atom):
            if not i.startswith("_"):
                self.insert(name=i, atom=getattr(xcffib.xproto.Atom, i))

    def insert(self, name=None, atom=None):
        assert name or atom
        if atom is None:
            c = self.conn.conn.core.InternAtom(False, len(name), name)
            atom = c.reply().atom
        if name is None:
            c = self.conn.conn.core.GetAtomName(atom)
            name = c.reply().name.to_string()
        self.atoms[name] = atom
        self.reverse[atom] = name

    def get_name(self, atom):
        if atom not in self.reverse:
            self.insert(atom=atom)
        return self.reverse[atom]

    def __getitem__(self, key):
        if key not in self.atoms:
            self.insert(name=key)
        return self.atoms[key]


class Screen:
    """
    This represents an actual X screen.
    """

    def __init__(self, conn, screen: xcffib.xproto.SCREEN):
        self.x_screen: xcffib.xproto.SCREEN = screen
        self.default_colormap = Colormap(conn, screen.default_colormap)
        self.root = window.XWindow(conn, self.x_screen.root)
        self.root_depth = self.x_screen.root_depth
        self.allowed_depths = self.x_screen.allowed_depths
        self.root_visual = self.x_screen.root_visual
        self.black_pixel = self.x_screen.black_pixel
        self.white_pixel = self.x_screen.white_pixel

        self._visuals: dict[int, xcffib.xproto.VISUALTYPE] = {}

        # Get visuals for 32 and 24 bit
        for d in [32, 24, self.root_depth]:
            if d not in self._visuals:
                visual = self.get_visual_for_depth(self, d)
                if visual:
                    self._visuals[d] = visual

    def _get_depth_and_visual(self, desired_depth):
        """
        Returns a tuple of (depth, visual) for the requested
        depth.

        Falls back to the root depth and visual if the requested
        depth is unavailable.
        """
        if desired_depth in self._visuals:
            return desired_depth, self._visuals[desired_depth]

        logger.info(
            "%s bit colour depth not available. Falling back to root depth: %s.",
            desired_depth,
            self.root_depth,
        )
        return self.root_depth, self._visuals[self.root_depth]

    @staticmethod
    def get_visual_for_depth(screen, depth):
        """
        Returns the visual object of the screen @ some depth

        For an ARGB visual -> depth=32
        For a RGB visual   -> depth=24
        """
        allowed = screen.allowed_depths
        if depth not in [x.depth for x in allowed]:
            logger.warning("Unsupported colour depth: %s", depth)
            return

        for i in allowed:
            if i.depth == depth:
                if i.visuals:
                    return i.visuals[0]


class Colormap:
    def __init__(self, conn, cid):
        self.conn = conn
        self.cid = cid

    def alloc_color(self, color):
        """
        Flexible color allocation.
        """
        try:
            return self.conn.conn.core.AllocNamedColor(self.cid, len(color), color).reply()
        except xcffib.xproto.NameError:

            def x8to16(i):
                return 0xFFFF * (i & 0xFF) // 0xFF

            try:
                color = hex(color)
            except ValueError:
                logger.error("Colormap failed to allocate %s", color)
                color = "#ff0000"

            r = x8to16(int(color[-6] + color[-5], 16))
            g = x8to16(int(color[-4] + color[-3], 16))
            b = x8to16(int(color[-2] + color[-1], 16))
            return self.conn.conn.core.AllocColor(self.cid, r, g, b).reply()


class Xinerama:
    def __init__(self, conn):
        self.ext = conn.conn(xcffib.xinerama.key)

    def query_screens(self):
        r = self.ext.QueryScreens().reply()
        return r.screen_info


class ScreenSaver:
    def __init__(self, conn):
        self.conn = conn
        self.ext = conn.conn(xcffib.screensaver.key)
        self.ext.QueryVersion(xcffib.screensaver.MAJOR_VERSION, xcffib.screensaver.MINOR_VERSION)

    def select_events(self, root):
        self.ext.SelectInput(
            root, xcffib.screensaver.Event.NotifyMask | xcffib.screensaver.Event.CycleMask
        )

    def set_interval(self, interval):
        self.conn.conn.core.SetScreenSaver(interval, 60, False, False)
        self.conn.conn.flush()


class RandR:
    def __init__(self, conn):
        self.ext = conn.conn(xcffib.randr.key)
        self.conn = conn

    def query_crtcs(self, root: int) -> list[Output]:
        infos: list[Output] = []
        primary = self.ext.GetOutputPrimary(root).reply().output
        for output in self.ext.GetScreenResources(root).reply().outputs:
            info = self.ext.GetOutputInfo(output, xcffib.CurrentTime).reply()

            # ignore disconnected monitors
            if info.connection != xcffib.randr.Connection.Connected:
                continue

            # ignore outputs with no monitor plugged in
            if not info.crtc:
                continue

            crtc_info = self.ext.GetCrtcInfo(info.crtc, xcffib.CurrentTime).reply()

            edid_raw = (
                self.ext.GetOutputProperty(
                    crtc_info.outputs[0], self.conn.atoms["EDID"], 0, 0, 256, False, False
                )
                .reply()
                .data
            )
            make = None
            model = None
            serial = None
            if len(edid_raw) > 0:
                (make, model, serial) = parse_edid(bytes(edid_raw))

            port = bytes(info.name).decode() if info.name else None
            rect = ScreenRect(crtc_info.x, crtc_info.y, crtc_info.width, crtc_info.height)
            out = Output(port, make, model, serial, rect)

            # prepend the primary output, append all others in screen
            # resources order
            if primary == output:
                infos.insert(0, out)
            else:
                infos.append(out)
        return infos

    def enable_screen_change_notifications(self, conn):
        self.ext.SelectInput(conn.default_screen.root.wid, xcffib.randr.NotifyMask.ScreenChange)


class XFixes:
    selection_mask = (
        SelectionEventMask.SetSelectionOwner
        | SelectionEventMask.SelectionClientClose
        | SelectionEventMask.SelectionWindowDestroy
    )

    def __init__(self, conn):
        self.conn = conn
        self.ext = conn.conn(xcffib.xfixes.key)
        self.ext.QueryVersion(xcffib.xfixes.MAJOR_VERSION, xcffib.xfixes.MINOR_VERSION)

    def select_selection_input(self, window, selection="PRIMARY"):
        _selection = self.conn.atoms[selection]
        self.conn.xfixes.ext.SelectSelectionInput(window.wid, _selection, self.selection_mask)


class Connection:
    _extmap = {
        "xinerama": Xinerama,
        "randr": RandR,
        "xfixes": XFixes,
        "mit-screen-saver": ScreenSaver,
    }

    def __init__(self, display):
        self.conn = xcffib.connect(display=display)
        self._connected = True
        self.cursors = Cursors(self)
        self.setup = self.conn.get_setup()
        extensions = self.extensions()
        self.screens = [Screen(self, i) for i in self.setup.roots]
        self.default_screen = self.screens[self.conn.pref_screen]

        for i in extensions:
            if i in self._extmap:
                setattr(self, i.replace("-", "_"), self._extmap[i](self))

        self.atoms = AtomCache(self)

        self.code_to_syms = {}
        self.sym_to_codes = None
        self.refresh_keymap()

        self.modmap = None
        self.refresh_modmap()

        self._cmaps = {}

    def colormap(self, desired_depth):
        if desired_depth in self._cmaps:
            return self._cmaps[desired_depth]

        _, visual = self.default_screen._get_depth_and_visual(desired_depth)

        cmap = self.conn.generate_id()
        self.conn.core.CreateColormap(
            xcffib.xproto.ColormapAlloc._None,
            cmap,
            self.default_screen.root.wid,
            visual.visual_id,
            is_checked=True,
        ).check()

        self._cmaps[desired_depth] = cmap
        return cmap

    @property
    def pseudoscreens(self):
        if hasattr(self, "randr"):
            return self.randr.query_crtcs(self.screens[0].root.wid)
        elif hasattr(self, "xinerama"):
            pseudoscreens = []
            for i, s in enumerate(self.xinerama.query_screens()):
                rect = ScreenRect(
                    s.x_org,
                    s.y_org,
                    s.width,
                    s.height,
                )
                pseudoscreens.append(Output(None, None, None, None, rect))
            return pseudoscreens
        raise Exception("no randr or xinerama?")

    def enable_screen_change_notifications(self):
        if not hasattr(self, "randr"):
            logger.warning(
                "no randr configured for this X server, screen change notifications disabled"
            )
            return
        self.randr.enable_screen_change_notifications(self)

    def finalize(self):
        self.cursors.finalize()
        self.disconnect()

    def refresh_keymap(self, first=None, count=None):
        if first is None:
            first = self.setup.min_keycode
            count = self.setup.max_keycode - self.setup.min_keycode + 1
        q = self.conn.core.GetKeyboardMapping(first, count).reply()

        assert len(q.keysyms) % q.keysyms_per_keycode == 0
        for i in range(len(q.keysyms) // q.keysyms_per_keycode):
            self.code_to_syms[first + i] = q.keysyms[
                i * q.keysyms_per_keycode : (i + 1) * q.keysyms_per_keycode
            ]

        sym_to_codes = {}
        for k, s in self.code_to_syms.items():
            for sym in s:
                if sym == 0:
                    continue
                if sym not in sym_to_codes:
                    sym_to_codes[sym] = [k]
                elif k not in sym_to_codes[sym]:
                    sym_to_codes[sym].append(k)

        self.sym_to_codes = sym_to_codes

    def refresh_modmap(self):
        reply = self.conn.core.GetModifierMapping().reply()
        modmap = {}
        names = (repeat(name, reply.keycodes_per_modifier) for name in ModMasks)
        for name, keycode in zip(chain.from_iterable(names), reply.keycodes):
            value = modmap.setdefault(name, [])
            value.append(keycode)
        self.modmap = modmap

    def get_modifier(self, keycode):
        """Return the modifier matching keycode"""
        for n, l in self.modmap.items():
            if keycode in l:
                return n
        return None

    def keysym_to_keycode(self, keysym):
        return self.sym_to_codes.get(keysym, [0])

    def keycode_to_keysym(self, keycode, modifier):
        if keycode >= len(self.code_to_syms) or modifier >= len(self.code_to_syms[keycode]):
            return 0
        return self.code_to_syms[keycode][modifier]

    def create_window(self, x, y, width, height, desired_depth=32):
        depth, visual = self.default_screen._get_depth_and_visual(desired_depth)

        wid = self.conn.generate_id()

        value_mask = CW.BackPixmap | CW.BorderPixel | CW.EventMask | CW.Colormap
        values = [
            xcffib.xproto.BackPixmap._None,
            0,
            EventMask.StructureNotify | EventMask.Exposure,
            self.colormap(depth),
        ]

        self.conn.core.CreateWindow(
            depth,
            wid,
            self.default_screen.root.wid,
            x,
            y,
            width,
            height,
            0,
            WindowClass.InputOutput,
            visual.visual_id,
            value_mask,
            values,
        )
        return window.XWindow(self, wid)

    def disconnect(self):
        with contextlib.suppress(xcffib.ConnectionException):
            self.conn.disconnect()
        self._connected = False

    def flush(self):
        if self._connected:
            return self.conn.flush()

    def xsync(self):
        # The idea here is that pushing an innocuous request through the queue
        # and waiting for a response "syncs" the connection, since requests are
        # serviced in order.
        self.conn.core.GetInputFocus().reply()

    def get_setup(self):
        return self.conn.get_setup()

    def extensions(self):
        return set(
            i.name.to_string().lower() for i in self.conn.core.ListExtensions().reply().names
        )

    def fixup_focus(self):
        """
        If the X11 focus is set to None, all keypress events are discarded,
        which makes our hotkeys not work. This fixes up the focus so it is not
        None.
        """
        window = self.conn.core.GetInputFocus().reply().focus
        if window == xcffib.xproto.InputFocus._None:
            self.conn.core.SetInputFocus(
                xcffib.xproto.InputFocus.PointerRoot,
                xcffib.xproto.InputFocus.PointerRoot,
                xcffib.xproto.Time.CurrentTime,
            )

    @functools.lru_cache
    def color_pixel(self, name):
        pixel = self.screens[0].default_colormap.alloc_color(name).pixel
        return pixel | 0xFF << 24


class Painter:
    def __init__(self, display):
        self.conn = xcffib.connect(display=display)
        self.setup = self.conn.get_setup()
        self.screens: list[Screen] = [Screen(self, i) for i in self.setup.roots]
        self.default_screen: Screen = self.screens[self.conn.pref_screen]
        self.conn.core.SetCloseDownMode(xcffib.xproto.CloseDown.RetainPermanent)
        self.atoms = AtomCache(self)
        self.width = -1
        self.height = -1
        self.root_pixmap_id = None

    def _get_root_pixmap_and_surface(self, screen) -> tuple[int, cairocffi.xcb.XCBSurface]:
        # Querying the screen dimensions via the xcffib connection does not
        # take account of any screen scaling. We can therefore work out the
        # necessary size of the root window by looking at the
        # pseudoscreens attribute and calculating the max x and y extents.
        root_windows = screen.qtile.core.conn.pseudoscreens
        width = max(win.rect.x + win.rect.width for win in root_windows)
        height = max(win.rect.y + win.rect.height for win in root_windows)

        try:
            root_pixmap = self.default_screen.root.get_property(
                "_XROOTPMAP_ID", xcffib.xproto.Atom.PIXMAP, int
            )
        except xcffib.ConnectionException:
            root_pixmap = None

        if not root_pixmap:
            root_pixmap = self.default_screen.root.get_property(
                "ESETROOT_PMAP_ID", xcffib.xproto.Atom.PIXMAP, int
            )
        if root_pixmap and (self.width == width and self.height == height):
            root_pixmap = root_pixmap[0]
        else:
            self.width = width
            self.height = height
            root_pixmap = self.conn.generate_id()
            self.conn.core.CreatePixmap(
                self.default_screen.root_depth,
                root_pixmap,
                self.default_screen.root.wid,
                self.width,
                self.height,
            )

        for depth in self.default_screen.allowed_depths:
            for visual in depth.visuals:
                if visual.visual_id == self.default_screen.root_visual:
                    root_visual = visual
                    break

        surface = cairocffi.xcb.XCBSurface(
            self.conn, root_pixmap, root_visual, self.width, self.height
        )

        return root_pixmap, surface

    def _update_root_pixmap(self, root_pixmap):
        self.conn.core.ChangeProperty(
            xcffib.xproto.PropMode.Replace,
            self.default_screen.root.wid,
            self.atoms["_XROOTPMAP_ID"],
            xcffib.xproto.Atom.PIXMAP,
            32,
            1,
            [root_pixmap],
        )
        self.conn.core.ChangeProperty(
            xcffib.xproto.PropMode.Replace,
            self.default_screen.root.wid,
            self.atoms["ESETROOT_PMAP_ID"],
            xcffib.xproto.Atom.PIXMAP,
            32,
            1,
            [root_pixmap],
        )
        self.conn.core.ChangeWindowAttributes(
            self.default_screen.root.wid, CW.BackPixmap, [root_pixmap]
        )
        self.conn.core.ClearArea(0, self.default_screen.root.wid, 0, 0, self.width, self.height)
        self.conn.flush()

        # now that we have drawn the new pixmap, free the old one
        if self.root_pixmap_id is not None and self.root_pixmap_id != root_pixmap:
            self.conn.core.FreePixmap(self.root_pixmap_id)
        self.root_pixmap_id = root_pixmap

    def fill(self, screen, background):
        root_pixmap, surface = self._get_root_pixmap_and_surface(screen)

        with cairocffi.Context(surface) as ctx:
            ctx.translate(screen.x, screen.y)
            ctx.rectangle(0, 0, screen.width, screen.height)
            ctx.set_source_rgba(*rgb(background))
            ctx.fill()

        surface.finish()
        self._update_root_pixmap(root_pixmap)

    def paint(self, screen, image_path, mode=None):
        try:
            with open(image_path, "rb") as f:
                image, _ = cairocffi.pixbuf.decode_to_image_surface(f.read())
        except OSError:
            logger.exception("Could not load wallpaper:")
            return

        root_pixmap, surface = self._get_root_pixmap_and_surface(screen)

        context = cairocffi.Context(surface)
        with context:
            context.translate(screen.x, screen.y)
            if mode == "fill":
                context.rectangle(0, 0, screen.width, screen.height)
                context.clip()
                image_w = image.get_width()
                image_h = image.get_height()
                width_ratio = screen.width / image_w
                if width_ratio * image_h >= screen.height:
                    context.scale(width_ratio)
                else:
                    height_ratio = screen.height / image_h
                    context.translate(-(image_w * height_ratio - screen.width) // 2, 0)
                    context.scale(height_ratio)
                context.set_source_surface(image)
            elif mode == "stretch":
                context.scale(
                    sx=screen.width / image.get_width(),
                    sy=screen.height / image.get_height(),
                )
                context.set_source_surface(image)
            elif mode == "center":
                target_x = (screen.width - image.get_width()) / 2
                target_y = (screen.height - image.get_height()) / 2
                context.set_source_surface(image, x=target_x, y=target_y)
            else:
                context.set_source_surface(image)
            context.paint()

        surface.finish()
        self._update_root_pixmap(root_pixmap)

    def __del__(self):
        self.conn.disconnect()


def get_keysym(key: str) -> int:
    keysym = keysyms.get(key.lower())
    if not keysym:
        raise XCBQError(f"Unknown key: {key}")
    return keysym


def translate_modifiers(mask: int) -> list[str]:
    r = []
    for k, v in ModMasks.items():
        if mask & v:
            r.append(k)
    return r


def translate_masks(modifiers: list[str]) -> int:
    """
    Translate a modifier mask specified as a list of strings into an or-ed
    bit representation.
    """
    masks = []
    for i in modifiers:
        try:
            masks.append(ModMasks[i.lower()])
        except KeyError as e:
            raise XCBQError(f"Unknown modifier: {i}") from e
    if masks:
        return functools.reduce(operator.or_, masks)
    else:
        return 0



from libqtile.widget.base import _TextBox


class ConfigErrorWidget(_TextBox):
    def __init__(self, **config):
        _TextBox.__init__(self, **config)
        self.class_name = self.widget.__class__.__name__
        self.text = f"Widget crashed: {self.class_name} (click to hide)"
        self.add_callbacks({"Button1": self._hide})

    def _hide(self):
        self.text = ""
        self.bar.draw()



# This module is used to allow the documentation to build properly
# when the ffi module cannot be imported:

# ../libqtile/backend/wayland/window.py:254: in <module>
#    @ffi.def_extern()
#     ^^^^^^^^^^^^^^
# E   AttributeError: 'NoneType' object has no attribute 'def_extern'

from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, TypeVar

if TYPE_CHECKING:
    from typing import Any

    T = TypeVar("T")


class FFIStub:
    @staticmethod
    def def_extern() -> Callable[[T], T]:
        return lambda f: f

    def __getattr__(self, name: str) -> Any:
        return None


class LibStub:
    def __getattr__(self, name: str) -> Any:
        # Return a sentinel int so enums work.
        return -1


ffi = FFIStub()
lib = LibStub()



from urllib.parse import quote, urlencode

from libqtile.widget import GenPollUrl


class Wttr(GenPollUrl):
    """Display weather widget provided by wttr.in_.

    .. _wttr.in: https://github.com/chubin/wttr.in/

    To specify your own custom output format, use the special %-notation
    (example: 'My_city: %t(%f), wind: %w'):

        - %c    Weather condition,
        - %C    Weather condition textual name,
        - %h    Humidity,
        - %t    Temperature (Actual),
        - %f    Temperature (Feels Like),
        - %w    Wind,
        - %l    Location,
        - %m    Moonphase 🌑🌒🌓🌔🌕🌖🌗🌘,
        - %M    Moonday,
        - %p    precipitation (mm),
        - %P    pressure (hPa),
        - %D    Dawn !,
        - %S    Sunrise !,
        - %z    Zenith !,
        - %s    Sunset !,
        - %d    Dusk !. (!times are shown in the local timezone)

    Add the character ``~`` at the beginning to get weather for some special
    location: ``~Vostok Station`` or ``~Eiffel Tower``.

    Also can use IP-addresses (direct) or domain names (prefixed with @) to
    specify a location:
    ``@github.com``, ``123.456.678.123``

    Specify multiple locations as dictionary ::

        location={
            'Minsk': 'Minsk',
            '64.127146,-21.873472': 'Reykjavik',
        }

    Cities will change randomly every update.
    """

    defaults = [
        (
            "format",
            "3",
            'Display text format. Choose presets in range 1-4 (Ex. ``"1"``) '
            "or build your own custom output format, use the special "
            "%-notation. See https://github.com/chubin/wttr.in#one-line-output",
        ),
        ("json", False, "Is Json?"),
        (
            "lang",
            "en",
            "Display text language. List of supported languages https://wttr.in/:translation",
        ),
        (
            "location",
            {},
            "Dictionary. Key is a city or place name, or GPS coordinates. "
            "Value is a display name. If the dictionary is empty, "
            "the location will be determined based on your IP address.",
        ),
        (
            "units",
            "m",
            "``'m'`` - metric, ``'M'`` - show wind speed in m/s, ``'u'`` - United States units",
        ),
        (
            "update_interval",
            600,
            "Update interval in seconds. Recommendation: if you want to "
            "display multiple locations alternately, maybe set a smaller "
            "interval, ex. ``30``.",
        ),
    ]

    def __init__(self, **config):
        GenPollUrl.__init__(self, json=False, **config)
        self.add_defaults(Wttr.defaults)
        self.url = self._get_url()

    def _get_url(self):
        params = {
            "format": self.format,
            "lang": self.lang,
        }
        location = ":".join(quote(loc) for loc in self.location)
        url = f"https://wttr.in/{location}?{self.units}&{urlencode(params)}"
        return url

    def parse(self, response):
        for coord in self.location:
            response = response.strip().replace(coord, self.location[coord])
        return response



from __future__ import annotations

import contextlib

import cairocffi
import xcffib.xproto

from libqtile import utils
from libqtile.backend.base import drawer
from libqtile.backend.x11 import window, xcbq


class Drawer(drawer.Drawer):
    """A helper class for drawing to Internal windows.

    The underlying surface here is an XCBSurface backed by a pixmap. We draw to the
    pixmap starting at offset 0, 0, and when the time comes to display to the window (on
    draw()), we copy the appropriate portion of the pixmap onto the window. In the event
    that our drawing area is resized, we invalidate the underlying surface and pixmap
    and recreate them when we need them again with the new geometry.
    """

    def __init__(self, conn: xcbq.Connection, win: window.Internal, width: int, height: int):
        drawer.Drawer.__init__(self, win, width, height)
        self.conn = conn
        self._xcb_surface = None
        self._gc = None
        self._depth, self._visual = conn.default_screen._get_depth_and_visual(win._depth)
        # Create an XCBSurface and pixmap
        self._check_xcb()

    def finalize(self):
        self._free_xcb_surface()
        self._free_pixmap()
        self._free_gc()
        drawer.Drawer.finalize(self)

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, width):
        if width > self._width:
            self._free_xcb_surface()
            self._free_pixmap()
        self._width = width

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, height):
        if height > self._height:
            self._free_xcb_surface()
            self._free_pixmap()
        self._height = height

    @property
    def pixmap(self):
        if self._pixmap is None:
            # draw here since the only use case of this function is in the
            # systray widget which expects a filled pixmap.
            self.draw()
        return self._pixmap

    def _create_gc(self):
        gc = self.conn.conn.generate_id()
        self.conn.conn.core.CreateGC(
            gc,
            self._win.wid,
            xcffib.xproto.GC.Foreground | xcffib.xproto.GC.Background,
            [
                self.conn.default_screen.black_pixel,
                self.conn.default_screen.white_pixel,
            ],
        )
        return gc

    def _free_gc(self):
        if self._gc is not None:
            with contextlib.suppress(xcffib.ConnectionException):
                self.conn.conn.core.FreeGC(self._gc)
            self._gc = None

    def _create_xcb_surface(self):
        surface = cairocffi.XCBSurface(
            self.conn.conn,
            self._pixmap,
            self._visual,
            self.width,
            self.height,
        )
        return surface

    def _free_xcb_surface(self):
        if self._xcb_surface is not None:
            self._xcb_surface.finish()
            self._xcb_surface = None

    def _create_pixmap(self):
        pixmap = self.conn.conn.generate_id()
        self.conn.conn.core.CreatePixmap(
            self._depth,
            pixmap,
            self._win.wid,
            self.width,
            self.height,
        )
        return pixmap

    def _free_pixmap(self):
        if self._pixmap is not None:
            with contextlib.suppress(xcffib.ConnectionException):
                self.conn.conn.core.FreePixmap(self._pixmap)
            self._pixmap = None

    def _check_xcb(self):
        # If the Drawer has been resized/invalidated we need to recreate these
        if self._xcb_surface is None:
            self._pixmap = self._create_pixmap()
            self._xcb_surface = self._create_xcb_surface()

    def _paint(self):
        # Paint RecordingSurface operations to the XCBSurface
        ctx = cairocffi.Context(self._xcb_surface)
        ctx.set_source_surface(self.surface, 0, 0)
        ctx.paint()

    def _draw(
        self,
        offsetx: int = 0,
        offsety: int = 0,
        width: int | None = None,
        height: int | None = None,
        src_x: int = 0,
        src_y: int = 0,
    ):
        # If this is our first draw, create the gc
        if self._gc is None:
            self._gc = self._create_gc()

        # Recreate an XCBSurface
        self._check_xcb()

        # paint stored operations(if any) to XCBSurface
        self._paint()

        # Finally, copy XCBSurface's underlying pixmap to the window.
        self.conn.conn.core.CopyArea(
            self._pixmap,
            self._win.wid,
            self._gc,
            src_x,
            src_y,  # srcx, srcy
            offsetx,
            offsety,  # dstx, dsty
            self.width if width is None else width,
            self.height if height is None else height,
        )

    def _find_root_visual(self):
        for i in self.conn.default_screen.allowed_depths:
            for v in i.visuals:
                if v.visual_id == self.conn.default_screen.root_visual:
                    return v

    def set_source_rgb(self, colour, ctx=None):
        # Remove transparency from non-32 bit windows
        if utils.has_transparency(colour) and self._depth != 32:
            colour = utils.remove_transparency(colour)

        drawer.Drawer.set_source_rgb(self, colour, ctx)

    def clear_rect(self, x=0, y=0, width=0, height=0):
        """
        Erases the background area specified by parameters. By default,
        the whole Drawer is cleared.

        The ability to clear a smaller area may be useful when you want to
        erase a smaller area of the drawer (e.g. drawing widget decorations).
        """
        if width <= 0:
            width = self.width
        if height <= 0:
            height = self.height

        self._check_xcb()

        # Using OPERATOR_CLEAR in a RecordingSurface does not clear the
        # XCBSurface so we clear the XCBSurface directly.
        with cairocffi.Context(self._xcb_surface) as ctx:
            ctx.set_operator(cairocffi.OPERATOR_CLEAR)
            ctx.rectangle(x, y, width, height)
            ctx.fill()



from libqtile.utils import lazify_imports

extensions = {
    "CommandSet": "command_set",
    "Dmenu": "dmenu",
    "DmenuRun": "dmenu",
    "J4DmenuDesktop": "dmenu",
    "RunCommand": "base",
    "WindowList": "window_list",
}

__all__, __dir__, __getattr__ = lazify_imports(extensions, __package__)



from libqtile import bar, hook, pangocffi
from libqtile.log_utils import logger
from libqtile.widget import base


class WindowTabs(base._TextBox):
    """
    Displays the name of each window in the current group.
    Contrary to TaskList this is not an interactive widget.
    The window that currently has focus is highlighted.
    """

    defaults = [
        ("separator", " | ", "Task separator text."),
        ("selected", ("<b>", "</b>"), "Selected task indicator"),
        (
            "parse_text",
            lambda window_name: window_name,
            "Function to modify window names. It must accept "
            "a string argument (original window name) and return "
            "a string with the modified name.",
        ),
    ]

    def __init__(self, **config):
        width = config.pop("width", bar.STRETCH)
        base._TextBox.__init__(self, width=width, **config)
        self.add_defaults(WindowTabs.defaults)
        if not isinstance(self.selected, tuple | list):
            self.selected = (self.selected, self.selected)

    def _configure(self, qtile, bar):
        base._TextBox._configure(self, qtile, bar)
        hook.subscribe.client_name_updated(self.update)
        hook.subscribe.focus_change(self.update)
        hook.subscribe.float_change(self.update)
        self.add_callbacks({"Button1": self.bar.screen.group.next_window})

    def update(self, *args):
        names = []
        for w in self.bar.screen.group.windows:
            try:
                name = self.parse_text(w.name if w and w.name else "")
            except:  # noqa: E722
                logger.exception("parse_text function failed:")
                name = w.name if w and w.name else "(unnamed)"
            state = ""
            if w.maximized:
                state = "[] "
            elif w.minimized:
                state = "_ "
            elif w.floating:
                state = "V "
            task = pangocffi.markup_escape_text(state + name)
            if w is self.bar.screen.group.current_window:
                task = task.join(self.selected)
            names.append(task)
        self.text = self.separator.join(names)
        self.bar.draw()



import itertools
import operator
import time
from os import statvfs

import cairocffi
import psutil

from libqtile.log_utils import logger
from libqtile.widget import base

__all__ = [
    "CPUGraph",
    "MemoryGraph",
    "SwapGraph",
    "NetGraph",
    "HDDGraph",
    "HDDBusyGraph",
]


class _Graph(base._Widget, base.MarginMixin):
    fixed_upper_bound = False
    defaults = [
        ("graph_color", "18BAEB", "Graph color"),
        ("fill_color", "1667EB.3", "Fill color for linefill graph"),
        ("border_color", "215578", "Widget border color"),
        ("border_width", 2, "Widget border width"),
        ("samples", 100, "Count of graph samples."),
        ("frequency", 1, "Update frequency in seconds"),
        ("type", "linefill", "'box', 'line', 'linefill'"),
        ("line_width", 3, "Line width"),
        ("start_pos", "bottom", "Drawer starting position ('bottom'/'top')"),
    ]

    def __init__(self, width=100, **config):
        base._Widget.__init__(self, width, **config)
        self.add_defaults(_Graph.defaults)
        self.values = [0] * self.samples
        self.maxvalue = 0
        self.oldtime = time.time()
        self.lag_cycles = 0

    def _configure(self, qtile, bar):
        super()._configure(qtile, bar)
        if self.type == "box":
            self.drawer.ctx.set_antialias(cairocffi.ANTIALIAS_NONE)

    def timer_setup(self):
        self.timeout_add(self.frequency, self.update)

    @property
    def graphwidth(self):
        return self.width - self.margin_x * 2 - self.border_width * 2

    @property
    def graphheight(self):
        return self.height - self.margin_y * 2 - self.border_width * 2

    def step(self):
        return self.graphwidth / float(self.samples)

    def _for_each_step(self, values):
        yield from enumerate(
            itertools.islice(
                values,
                max(int(-(self.graphwidth / self.step()) + len(values)), 0),
                len(values),
            )
        )

    def _prepare_context(self):
        self.drawer.ctx.set_line_join(cairocffi.LINE_JOIN_ROUND)
        if self.graph_color is not None:
            self.drawer.set_source_rgb(self.graph_color)
        self.drawer.ctx.set_line_width(self.line_width)

    def draw_box(self, x, y, values):
        self._prepare_context()
        for _, val in self._for_each_step(values):
            val = self.val(val)
            self.drawer.ctx.rectangle(x, y - val, self.step(), val)
            x += self.step()
        self.drawer.ctx.fill()
        self.drawer.ctx.stroke()

    def draw_line(self, x, y, values):
        self._prepare_context()
        for _, val in self._for_each_step(values):
            self.drawer.ctx.line_to(x, y - self.val(val))
            x += self.step()
        self.drawer.ctx.stroke()

    def draw_linefill(self, x, y, values):
        self._prepare_context()
        for index, val in self._for_each_step(values):
            self.drawer.ctx.line_to(x + index * self.step(), y - self.val(val))
        self.drawer.ctx.stroke_preserve()
        self.drawer.ctx.line_to(
            x + (len(values) - 1) * self.step(), y - 1 + self.line_width / 2.0
        )
        self.drawer.ctx.line_to(x, y - 1 + self.line_width / 2.0)
        self.drawer.set_source_rgb(self.fill_color)
        self.drawer.ctx.fill()

    def val(self, val):
        if self.start_pos == "bottom":
            return val
        elif self.start_pos == "top":
            return -val
        else:
            raise ValueError(f"Unknown starting position: {self.start_pos}.")

    def draw(self):
        self.drawer.clear(self.background or self.bar.background)
        if self.border_width:
            self.drawer.set_source_rgb(self.border_color)
            self.drawer.ctx.set_line_width(self.border_width)
            self.drawer.ctx.rectangle(
                self.margin_x + self.border_width / 2.0,
                self.margin_y + self.border_width / 2.0,
                self.graphwidth + self.border_width,
                self.graphheight + self.border_width,
            )
            self.drawer.ctx.stroke()
        x = self.margin_x + self.border_width
        y = self.margin_y + self.border_width
        if self.start_pos == "bottom":
            y += self.graphheight
        elif not self.start_pos == "top":
            raise ValueError(f"Unknown starting position: {self.start_pos}.")
        k = 1.0 / (self.maxvalue or 1)
        scaled = [self.graphheight * val * k for val in reversed(self.values)]

        if self.type == "box":
            self.draw_box(x, y, scaled)
        elif self.type == "line":
            self.draw_line(x, y, scaled)
        elif self.type == "linefill":
            self.draw_linefill(x, y, scaled)
        else:
            raise ValueError(f"Unknown graph type: {self.type}.")

        self.draw_at_default_position()

    def push(self, value):
        if self.lag_cycles > self.samples:
            # compensate lag by sending the same value up to
            # the graph samples limit
            self.lag_cycles = 1

        self.values = ([value] * min(self.samples, self.lag_cycles)) + self.values
        self.values = self.values[: self.samples]

        if not self.fixed_upper_bound:
            self.maxvalue = max(self.values)
        self.draw()

    def update(self):
        # lag detection
        newtime = time.time()
        self.lag_cycles = int((newtime - self.oldtime) / self.frequency)
        self.oldtime = newtime

        self.update_graph()
        self.timeout_add(self.frequency, self.update)

    def fulfill(self, value):
        self.values = [value] * len(self.values)


class CPUGraph(_Graph):
    """Display CPU usage graph.

    Widget requirements: psutil_.

    .. _psutil: https://pypi.org/project/psutil/
    """

    orientations = base.ORIENTATION_HORIZONTAL
    defaults = [
        ("core", "all", "Which core to show (all/0/1/2/...)"),
    ]

    fixed_upper_bound = True

    def __init__(self, **config):
        _Graph.__init__(self, **config)
        self.add_defaults(CPUGraph.defaults)
        self.maxvalue = 100
        self.oldvalues = self._getvalues()

    def _getvalues(self):
        if isinstance(self.core, int):
            if self.core > psutil.cpu_count() - 1:
                raise ValueError(f"No such core: {self.core}")
            cpu = psutil.cpu_times(percpu=True)[self.core]
        else:
            cpu = psutil.cpu_times()

        user = cpu.user * 100
        nice = cpu.nice * 100
        sys = cpu.system * 100
        idle = cpu.idle * 100

        return (int(user), int(nice), int(sys), int(idle))

    def update_graph(self):
        nval = self._getvalues()
        oval = self.oldvalues
        busy = nval[0] + nval[1] + nval[2] - oval[0] - oval[1] - oval[2]
        total = busy + nval[3] - oval[3]
        # sometimes this value is zero for unknown reason (time shift?)
        # we just sent the previous value, because it gives us no info about
        # cpu load, if it's zero.

        if total:
            push_value = busy * 100.0 / total
            self.push(push_value)
        else:
            self.push(self.values[0])
        self.oldvalues = nval


class MemoryGraph(_Graph):
    """Displays a memory usage graph.

    Widget requirements: psutil_.

    .. _psutil: https://pypi.org/project/psutil/
    """

    orientations = base.ORIENTATION_HORIZONTAL
    fixed_upper_bound = True

    def __init__(self, **config):
        _Graph.__init__(self, **config)
        val = self._getvalues()
        self.maxvalue = val["MemTotal"]

        mem = val["MemTotal"] - val["MemFree"] - val["Buffers"] - val["Cached"]
        self.fulfill(mem)

    def _getvalues(self):
        val = {}
        mem = psutil.virtual_memory()
        val["MemTotal"] = int(mem.total / 1024 / 1024)
        val["MemFree"] = int(mem.free / 1024 / 1024)
        val["Buffers"] = int(mem.buffers / 1024 / 1024)
        val["Cached"] = int(mem.cached / 1024 / 1024)
        return val

    def update_graph(self):
        val = self._getvalues()
        self.push(val["MemTotal"] - val["MemFree"] - val["Buffers"] - val["Cached"])


class SwapGraph(_Graph):
    """Display a swap info graph.

    Widget requirements: psutil_.

    .. _psutil: https://pypi.org/project/psutil/
    """

    orientations = base.ORIENTATION_HORIZONTAL
    fixed_upper_bound = True

    def __init__(self, **config):
        _Graph.__init__(self, **config)
        val = self._getvalues()
        self.maxvalue = val["SwapTotal"]
        swap = val["SwapTotal"] - val["SwapFree"]
        self.fulfill(swap)

    def _getvalues(self):
        val = {}
        swap = psutil.swap_memory()
        val["SwapTotal"] = int(swap.total / 1024 / 1024)
        val["SwapFree"] = int(swap.free / 1024 / 1024)
        return val

    def update_graph(self):
        val = self._getvalues()

        swap = val["SwapTotal"] - val["SwapFree"]

        # can change, swapon/off
        if self.maxvalue != val["SwapTotal"]:
            self.maxvalue = val["SwapTotal"]
            self.fulfill(swap)
        self.push(swap)


class NetGraph(_Graph):
    """Display a network usage graph.

    Widget requirements: psutil_.

    .. _psutil: https://pypi.org/project/psutil/"""

    orientations = base.ORIENTATION_HORIZONTAL
    defaults = [
        ("interface", "auto", "Interface to display info for ('auto' for detection)"),
        ("bandwidth_type", "down", "down(load)/up(load)"),
    ]

    def __init__(self, **config):
        _Graph.__init__(self, **config)
        self.add_defaults(NetGraph.defaults)
        if self.interface == "auto":
            try:
                self.interface = self.get_main_iface()
            except RuntimeError:
                logger.warning(
                    "NetGraph - Automatic interface detection failed, falling back to 'eth0'"
                )
                self.interface = "eth0"
        if self.bandwidth_type != "down" and self.bandwidth_type != "up":
            raise ValueError(f"bandwidth type {self.bandwidth_type} not known!")
        self.bytes = 0
        self.bytes = self._get_values()

    def _get_values(self):
        net = psutil.net_io_counters(pernic=True)
        if self.bandwidth_type == "up":
            return net[self.interface].bytes_sent
        if self.bandwidth_type == "down":
            return net[self.interface].bytes_recv

    def update_graph(self):
        val = self._get_values()
        change = val - self.bytes
        self.bytes = val
        self.push(change)

    @staticmethod
    def get_main_iface():
        # XXX: psutil doesn't have the facility to get the main interface,
        # so I'll just return the interface that has received the most traffic.
        #
        # I could do this with netifaces, but that's another dependency.
        #
        # Oh. and there is probably a better way to do this.

        net = psutil.net_io_counters(pernic=True)
        iface = {}
        for entry in net:
            iface[entry] = net[entry].bytes_recv
        return sorted(iface.items(), key=operator.itemgetter(1))[-1][0]


class HDDGraph(_Graph):
    """Display HDD free or used space graph"""

    fixed_upper_bound = True
    orientations = base.ORIENTATION_HORIZONTAL
    defaults = [("path", "/", "Partition mount point."), ("space_type", "used", "free/used")]

    def __init__(self, **config):
        _Graph.__init__(self, **config)
        self.add_defaults(HDDGraph.defaults)
        stats = statvfs(self.path)
        self.maxvalue = stats.f_blocks * stats.f_frsize
        values = self._get_values()
        self.fulfill(values)

    def _get_values(self):
        stats = statvfs(self.path)
        if self.space_type == "used":
            return (stats.f_blocks - stats.f_bfree) * stats.f_frsize
        else:
            return stats.f_bavail * stats.f_frsize

    def update_graph(self):
        val = self._get_values()
        self.push(val)


class HDDBusyGraph(_Graph):
    """Display HDD busy time graph

    Parses /sys/block/<dev>/stat file and extracts overall device IO usage,
    based on ``io_ticks``'s value.  See
    https://www.kernel.org/doc/Documentation/block/stat.txt
    """

    orientations = base.ORIENTATION_HORIZONTAL
    defaults = [("device", "sda", "Block device to display info for")]

    def __init__(self, **config):
        _Graph.__init__(self, **config)
        self.add_defaults(HDDBusyGraph.defaults)
        self.path = f"/sys/block/{self.device}/stat"
        self._prev = 0

    def _get_values(self):
        try:
            # io_ticks is field number 9
            with open(self.path) as f:
                io_ticks = int(f.read().split()[9])
        except OSError:
            return 0
        activity = io_ticks - self._prev
        self._prev = io_ticks
        return activity

    def update_graph(self):
        val = self._get_values()
        self.push(val)



from subprocess import check_output

from libqtile.lazy import lazy
from libqtile.log_utils import logger
from libqtile.widget import base


class DoNotDisturb(base.InLoopPollText):
    """
    Displays Do Not Disturb status for notification server Dunst by default.
    Can be used with other servers by changing the poll command and mouse callbacks.
    """

    defaults = [
        (
            "poll_function",
            None,
            "Function that returns the notification server status. "
            "Define the function on your configuration file and "
            "pass it like poll_function=my_func. "
            "Must return either true or false",
        ),
        ("enabled_icon", "X", "Icon that displays when do not disturb is enabled"),
        ("disabled_icon", "O", "Icon that displays when do not disturb is disabled"),
        ("update_interval", 1, "How often in seconds the text must update"),
    ]

    def __init__(self, **config):
        base.InLoopPollText.__init__(self, **config)
        self.add_defaults(DoNotDisturb.defaults)
        self.status_retrieved_error = False
        if self.poll_function is None:
            self.add_callbacks(
                {
                    "Button1": lazy.spawn("dunstctl set-paused toggle"),
                    "Button3": lazy.spawn("dunstctl history-pop"),
                }
            )

    def dunst_status(self):
        status = check_output(["dunstctl", "is-paused"]).strip()
        if status == b"true":
            return True
        return False

    def poll(self):
        check = None
        if self.poll_function is None:
            check = self.dunst_status()
        elif callable(self.poll_function):
            check = self.poll_function()
        else:
            if not self.status_retrieved_error:
                logger.error("Custom poll function cannot be called")
                self.status_retrieved_error = True
        if check:
            return self.enabled_icon
        else:
            return self.disabled_icon



from libqtile import bar, hook, pangocffi
from libqtile.backend.x11 import xcbq
from libqtile.widget import base


class Clipboard(base._TextBox):
    """Display current clipboard contents"""

    defaults = [
        ("selection", "CLIPBOARD", "the selection to display(CLIPBOARD or PRIMARY)"),
        (
            "max_width",
            10,
            "maximum number of characters to display "
            "(None for all, useful when width is bar.STRETCH)",
        ),
        ("timeout", 10, "Default timeout (seconds) for display text, None to keep forever"),
        (
            "blacklist",
            ["keepassx"],
            "list with blacklisted wm_class, sadly not every "
            "clipboard window sets them, keepassx does."
            "Clipboard contents from blacklisted wm_classes "
            "will be replaced by the value of ``blacklist_text``.",
        ),
        ("blacklist_text", "***********", "text to display when the wm_class is blacklisted"),
    ]

    def __init__(self, width=bar.CALCULATED, **config):
        base._TextBox.__init__(self, "", width, **config)
        self.add_defaults(Clipboard.defaults)
        self.timeout_id = None

    def _configure(self, qtile, bar):
        base._TextBox._configure(self, qtile, bar)
        self.text = ""
        self.setup_hooks()

    def clear(self, *args):
        self.text = ""
        self.bar.draw()

    def is_blacklisted(self, owner_id):
        if not self.blacklist:
            return False

        if owner_id in self.qtile.windows_map:
            owner = self.qtile.windows_map[owner_id].window
        else:
            owner = xcbq.window.XWindow(self.qtile.core.conn, owner_id)

        owner_class = owner.get_wm_class()
        if owner_class:
            for wm_class in self.blacklist:
                if wm_class in owner_class:
                    return True

    def setup_hooks(self):
        def hook_change(name, selection):
            if name != self.selection:
                return

            if self.is_blacklisted(selection["owner"]):
                text = self.blacklist_text
            else:
                text = selection["selection"].replace("\n", " ")

                text = text.strip()
                if self.max_width is not None and len(text) > self.max_width:
                    text = text[: self.max_width] + "..."

            self.text = pangocffi.markup_escape_text(text)

            if self.timeout_id:
                self.timeout_id.cancel()
                self.timeout_id = None

            if self.timeout:
                self.timeout_id = self.timeout_add(self.timeout, self.clear)
            self.bar.draw()

        def hook_notify(name, selection):
            if name != self.selection:
                return

            if self.timeout_id:
                self.timeout_id.cancel()
                self.timeout_id = None

            # only clear if don't change don't apply in .5 seconds
            if self.timeout:
                self.timeout_id = self.timeout_add(self.timeout, self.clear)
            self.bar.draw()

        hook.subscribe.selection_notify(hook_notify)
        hook.subscribe.selection_change(hook_change)



from __future__ import annotations

import asyncio
import copy
import inspect
import math
import subprocess
from typing import TYPE_CHECKING

from libqtile import bar, configurable, confreader, hook
from libqtile.command import interface
from libqtile.command.base import CommandObject, expose_command
from libqtile.lazy import LazyCall
from libqtile.log_utils import logger
from libqtile.utils import ColorType, create_task

if TYPE_CHECKING:
    from typing import Any

    from libqtile.command.base import ItemT

# Each widget class must define which bar orientation(s) it supports by setting
# these bits in an 'orientations' class attribute. Simply having the attribute
# inherited by superclasses is discouraged, because if a superclass that was
# only supporting one orientation, adds support for the other, its subclasses
# will have to be adapted too, in general. ORIENTATION_NONE is only added for
# completeness' sake.
# +------------------------+--------------------+--------------------+
# | Widget bits            | Horizontal bar     | Vertical bar       |
# +========================+====================+====================+
# | ORIENTATION_NONE       | ConfigError raised | ConfigError raised |
# +------------------------+--------------------+--------------------+
# | ORIENTATION_HORIZONTAL | Widget displayed   | ConfigError raised |
# |                        | horizontally       |                    |
# +------------------------+--------------------+--------------------+
# | ORIENTATION_VERTICAL   | ConfigError raised | Widget displayed   |
# |                        |                    | vertically         |
# +------------------------+--------------------+--------------------+
# | ORIENTATION_BOTH       | Widget displayed   | Widget displayed   |
# |                        | horizontally       | vertically         |
# +------------------------+--------------------+--------------------+


class _Orientations(int):
    def __new__(cls, value, doc):
        return super().__new__(cls, value)

    def __init__(self, value, doc):
        self.doc = doc

    def __str__(self):
        return self.doc

    def __repr__(self):
        return self.doc


ORIENTATION_NONE = _Orientations(0, "none")
ORIENTATION_HORIZONTAL = _Orientations(1, "horizontal only")
ORIENTATION_VERTICAL = _Orientations(2, "vertical only")
ORIENTATION_BOTH = _Orientations(3, "horizontal and vertical")


class _Widget(CommandObject, configurable.Configurable):
    """Base Widget class

    If length is set to the special value `bar.STRETCH`, the bar itself will
    set the length to the maximum remaining space, after all other widgets have
    been configured.

    In horizontal bars, 'length' corresponds to the width of the widget; in
    vertical bars, it corresponds to the widget's height.

    The offsetx and offsety attributes are set by the Bar after all widgets
    have been configured.

    Callback functions can be assigned to button presses by passing a dict to the
    'callbacks' kwarg. No arguments are passed to the function so, if
    you need access to the qtile object, it needs to be imported into your code.

    ``lazy`` functions can also be passed as callback functions and can be used in
    the same way as keybindings.

    For example:

    .. code-block:: python

        from libqtile import qtile

        def open_calendar():
            qtile.spawn('gsimplecal next_month')

        clock = widget.Clock(
            mouse_callbacks={
                'Button1': open_calendar,
                'Button3': lazy.spawn('gsimplecal prev_month')
            }
        )

    When the clock widget receives a click with button 1, the ``open_calendar`` function
    will be executed.
    """

    orientations = ORIENTATION_BOTH

    # Default (empty set) is for all backends to be supported. Widgets can override this
    # to explicitly confirm which backends are supported
    supported_backends: set[str] = set()

    offsetx: int = 0
    offsety: int = 0
    defaults: list[tuple[str, Any, str]] = [
        ("background", None, "Widget background color"),
        (
            "mouse_callbacks",
            {},
            "Dict of mouse button press callback functions. Accepts functions and ``lazy`` calls.",
        ),
        ("hide_crash", False, "Don't display error in bar if widget crashes on startup."),
    ]

    def __init__(self, length, **config):
        """
        length: bar.STRETCH, bar.CALCULATED, or a specified length.
        """
        CommandObject.__init__(self)
        self.name = self.__class__.__name__.lower()
        if "name" in config:
            self.name = config["name"]

        configurable.Configurable.__init__(self, **config)

        # Add defaults for Mixins if inherited
        if isinstance(self, PaddingMixin):
            self.add_defaults(PaddingMixin.defaults)
        if isinstance(self, MarginMixin):
            self.add_defaults(MarginMixin.defaults)

        self.add_defaults(_Widget.defaults)

        if length in (bar.CALCULATED, bar.STRETCH):
            self.length_type = length
            self.length = 0
        elif isinstance(length, int):
            self.length_type = bar.STATIC
            self.length = length
        else:
            raise confreader.ConfigError("Widget width must be an int")

        self.configured = False
        self._futures: list[asyncio.Handle] = []
        self._mirrors: set[_Widget] = set()
        self.finalized = False

    @property
    def length(self):
        if self.length_type == bar.CALCULATED:
            try:
                return int(self.calculate_length())
            except Exception:
                logger.exception(f"error when calculating widget {self.name} length")
                return 0
        return self._length

    @length.setter
    def length(self, value):
        self._length = value

    @property
    def width(self):
        if self.bar.horizontal:
            return self.length
        return self.bar.width

    @property
    def height(self):
        if self.bar.horizontal:
            return self.bar.height
        return self.length

    def _test_orientation_compatibility(self, horizontal):
        if horizontal:
            if not self.orientations & ORIENTATION_HORIZONTAL:
                raise confreader.ConfigError(
                    self.__class__.__name__
                    + " is not compatible with the orientation of the bar."
                )
        elif not self.orientations & ORIENTATION_VERTICAL:
            raise confreader.ConfigError(
                self.__class__.__name__ + " is not compatible with the orientation of the bar."
            )

    def timer_setup(self):
        """This is called exactly once, after the widget has been configured
        and timers are available to be set up."""

    def _configure(self, qtile, bar):
        self._test_orientation_compatibility(bar.horizontal)

        self.qtile = qtile
        self.bar = bar
        self.drawer = bar.window.create_drawer(self.bar.width, self.bar.height)

        # Clear this flag as widget may be restarted (e.g. if screen removed and re-added)
        self.finalized = False

        # Timers are added to futures list so they can be cancelled if the `finalize` method is
        # called before the timers have fired.
        if not self.configured:
            timer = self.qtile.call_soon(self.timer_setup)
            async_timer = self.qtile.call_soon(asyncio.create_task, self._config_async())

            # Add these to our list of futures so they can be cancelled.
            self._futures.extend([timer, async_timer])
            if hasattr(self, "force_update"):
                hook.subscribe.resume(self.force_update)

    async def _config_async(self):
        """
        This is called once when the main eventloop has started. this
        happens after _configure has been run.

        Widgets that need to use asyncio coroutines after this point may
        wish to initialise the relevant code (e.g. connections to dbus
        using dbus_fast) here.
        """

    def finalize(self):
        for future in self._futures:
            future.cancel()
        if hasattr(self, "layout") and self.layout:
            self.layout.finalize()
            self.layout = None
        self.drawer.finalize()
        self.finalized = True

        # Reset configuration status so the widget can be reconfigured
        # e.g. when screen is re-added
        self.configured = False

    def clear(self):
        self.drawer.set_source_rgb(self.bar.background)
        self.drawer.fillrect(self.offsetx, self.offsety, self.width, self.height)

    @expose_command()
    def info(self):
        """Info for this object."""
        return dict(
            name=self.name,
            offset=self.offsetx if self.bar.horizontal else self.offsety,
            length=self.length,
            width=self.width,
            height=self.height,
        )

    def add_callbacks(self, defaults):
        """Add default callbacks with a lower priority than user-specified callbacks."""
        defaults.update(self.mouse_callbacks)
        self.mouse_callbacks = defaults

    def button_press(self, x, y, button):
        name = f"Button{button}"
        if name in self.mouse_callbacks:
            cmd = self.mouse_callbacks[name]
            if isinstance(cmd, LazyCall):
                if cmd.check(self.qtile):
                    status, val = self.qtile.server.call(
                        (cmd.selectors, cmd.name, cmd.args, cmd.kwargs, False)
                    )
                    if status in (interface.ERROR, interface.EXCEPTION):
                        logger.error("Mouse callback command error %s: %s", cmd.name, val)
            else:
                cmd()

    def button_release(self, x, y, button):
        pass

    def _items(self, name: str) -> ItemT:
        if name == "bar":
            return True, []
        elif name == "screen":
            return True, []
        return None

    def _select(self, name, sel):
        if name == "bar":
            return self.bar
        elif name == "screen":
            return self.bar.screen

    def rotate_drawer_left(self):
        # Left bar reads bottom to top
        self.drawer.ctx.rotate(-90 * math.pi / 180.0)
        self.drawer.ctx.translate(-self.length, 0)

    def rotate_drawer_right(self):
        # Right bar is top to bottom
        self.drawer.ctx.translate(self.bar.width, 0)
        self.drawer.ctx.rotate(90 * math.pi / 180.0)

    def rotate_drawer(self):
        if self.bar.horizontal:
            return
        if self.bar.screen.left is self.bar:
            self.rotate_drawer_left()
        elif self.bar.screen.right is self.bar:
            self.rotate_drawer_right()

    def draw_at_default_position(self):
        """Default position to draw the widget in horizontal and vertical bars."""
        self.drawer.draw(
            offsetx=self.offsetx, offsety=self.offsety, width=self.width, height=self.height
        )

    def draw(self):
        """
        Method that draws the widget. You may call this explicitly to
        redraw the widget, but only if the length of the widget hasn't
        changed. If it has, you must call bar.draw instead.
        """
        raise NotImplementedError

    def calculate_length(self):
        """
        Must be implemented if the widget can take CALCULATED for length.
        It must return the width of the widget if it's installed in a
        horizontal bar; it must return the height of the widget if it's
        installed in a vertical bar. Usually you will test the orientation
        of the bar with 'self.bar.horizontal'.
        """
        raise NotImplementedError

    def timeout_add(self, seconds, method, method_args=()):
        """
        This method calls ``.call_later`` with given arguments.
        """
        # Don't add timers for finalised widgets
        if self.finalized:
            return

        future = self.qtile.call_later(seconds, self._wrapper, method, *method_args)

        self._futures.append(future)
        return future

    def call_process(self, command, **kwargs):
        """
        This method uses `subprocess.check_output` to run the given command
        and return the string from stdout, which is decoded when using
        Python 3.
        """
        return subprocess.check_output(command, **kwargs, encoding="utf-8")

    def _remove_dead_timers(self):
        """Remove completed and cancelled timers from the list."""

        def is_ready(timer):
            return timer in self.qtile._eventloop._ready

        self._futures = [
            timer
            for timer in self._futures
            # Filter out certain handles...
            if not (
                timer.cancelled()
                # Once a scheduled timer is ready to be run its _scheduled flag is set to False
                # and it's added to the loop's `_ready` queue
                or (
                    isinstance(timer, asyncio.TimerHandle)
                    and not timer._scheduled
                    and not is_ready(timer)
                )
                # Callbacks scheduled via `call_soon` are put into the loop's `_ready` queue
                # and are removed once they've been executed
                or (isinstance(timer, asyncio.Handle) and not is_ready(timer))
            )
        ]

    def _wrapper(self, method, *method_args):
        self._remove_dead_timers()
        try:
            if inspect.iscoroutinefunction(method):
                create_task(method(*method_args))
            elif asyncio.iscoroutine(method):
                create_task(method)
            else:
                method(*method_args)
        except:  # noqa: E722
            logger.exception("got exception from widget timer")

    def create_mirror(self):
        return Mirror(self, background=self.background)

    def clone(self):
        return copy.deepcopy(self)

    def mouse_enter(self, x, y):
        pass

    def mouse_leave(self, x, y):
        pass

    def _draw_with_mirrors(self) -> None:
        self._old_draw()
        for mirror in self._mirrors:
            if not mirror.configured:
                continue

            # If the widget and mirror are on the same bar then we could have an
            # infinite loop when we call bar.draw(). mirror.draw() will trigger a resize
            # if it's the wrong size.
            if mirror.length_type == bar.CALCULATED and mirror.bar is not self.bar:
                mirror.bar.draw()
            else:
                mirror.draw()

    def add_mirror(self, widget: _Widget):
        if not self._mirrors:
            self._old_draw = self.draw
            self.draw = self._draw_with_mirrors

        self._mirrors.add(widget)
        if not self.drawer.has_mirrors:
            self.drawer.has_mirrors = True

    def remove_mirror(self, widget: _Widget):
        try:
            self._mirrors.remove(widget)
        except KeyError:
            pass

        if not self._mirrors:
            self.drawer.has_mirrors = False

            if hasattr(self, "_old_draw"):
                # Deletes the reference to draw and falls back to the original
                del self.draw
                del self._old_draw


class _TextBox(_Widget):
    """
    Base class for widgets that are just boxes containing text.
    """

    orientations = ORIENTATION_BOTH
    defaults = [
        ("font", "sans", "Default font"),
        ("fontsize", None, "Font size. Calculated if None."),
        ("padding", None, "Padding. Calculated if None."),
        ("foreground", "ffffff", "Foreground colour"),
        ("fontshadow", None, "font shadow color, default is None(no shadow)"),
        ("markup", True, "Whether or not to use pango markup"),
        (
            "fmt",
            "{}",
            "Format to apply to the string returned by the widget. Main purpose: applying markup. "
            "For a widget that returns ``foo``, using ``fmt='<i>{}</i>'`` would give you ``<i>foo</i>``. "
            "To control what the widget outputs in the first place, use the ``format`` paramater of the widget (if it has one).",
        ),
        ("max_chars", 0, "Maximum number of characters to display in widget."),
        (
            "scroll",
            False,
            "Whether text should be scrolled. When True, you must set the widget's ``width``.",
        ),
        (
            "scroll_repeat",
            True,
            "Whether text should restart scrolling once the text has ended",
        ),
        (
            "scroll_delay",
            2,
            "Number of seconds to pause before starting scrolling and restarting/clearing text at end",
        ),
        ("scroll_step", 1, "Number of pixels to scroll with each step"),
        ("scroll_interval", 0.1, "Time in seconds before next scrolling step"),
        (
            "scroll_clear",
            False,
            "Whether text should scroll completely away (True) or stop when the end of the text is shown (False)",
        ),
        ("scroll_hide", False, "Whether the widget should hide when scrolling has finished"),
        (
            "scroll_fixed_width",
            False,
            "When ``scroll=True`` the ``width`` parameter is a maximum width and, when text is shorter than this, the widget will resize. "
            "Setting ``scroll_fixed_width=True`` will force the widget to have a fixed width, regardless of the size of the text.",
        ),
        ("rotate", True, "Rotate text in vertical bar."),
        (
            "direction",
            "default",
            "Override the text direction in vertical bar, has no effect on text in horizontal bar."
            "default: text displayed based on vertical bar position (left/right)"
            "ttb: text read from top to bottom, btt: text read from bottom to top."
            "'default', 'ttb', 'btt'",
        ),
    ]  # type: list[tuple[str, Any, str]]

    def __init__(self, text=" ", width=bar.CALCULATED, **config):
        self.layout = None
        _Widget.__init__(self, width, **config)
        self.add_defaults(_TextBox.defaults)
        self.text = text
        self._is_scrolling = False
        self._should_scroll = False
        self._scroll_offset = 0
        self._scroll_queued = False
        self._scroll_timer = None
        self._scroll_width = width

    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, value):
        if len(value) > self.max_chars > 0:
            value = value[: self.max_chars] + "…"
        self._text = value
        if self.layout:
            self.layout.text = self.formatted_text

            if self.scroll:
                self.check_width()
                self.reset_scroll()

    @property
    def formatted_text(self):
        return self.fmt.format(self._text)

    def _configure(self, qtile, bar):
        _Widget._configure(self, qtile, bar)
        if self.fontsize is None:
            self.fontsize = self.bar.size - self.bar.size / 5
        if self.padding is None:
            self.padding = self.fontsize // 2
        if self.direction not in ("default", "ttb", "btt"):
            logger.warning(
                "Invalid value set for direction: %s. Valid values are: 'default', 'ttb', 'btt'. "
                "direction has been set to 'default'",
                self.direction,
            )
            self.direction = "default"
        self.layout = self.drawer.textlayout(
            self.formatted_text,
            self.foreground,
            self.font,
            self.fontsize,
            self.fontshadow,
            markup=self.markup,
        )
        if not isinstance(self._scroll_width, int) and self.scroll:
            if not self.bar.horizontal and not self.rotate:
                self._scroll_width = self.bar.width
            else:
                logger.warning("%s: You must specify a width when enabling scrolling.", self.name)
                self.scroll = False

        # Setting the layout width will wrap text which increases layout's height,
        # we only want this when bar is vertical and rotation is disabled
        # to be able to display more of the text using multiple lines,
        # only if scrolling is enabled the layout width will be overwritten
        # because the widget's width is handle by scroll.
        if not self.bar.horizontal and not self.rotate:
            self.layout.width = self.bar.width

        if self.scroll:
            self.check_width()

    def check_width(self):
        """
        Check whether the widget needs to have calculated or fixed width
        and whether the text should be scrolled.
        """
        # Reset the layout width to let the layout calculate
        # the width based on the length of the text.
        self.layout.reset_width()
        if self.layout.width > self._scroll_width:
            if self.bar.horizontal or self.rotate:
                self.length_type = bar.STATIC
                self.length = self._scroll_width
            self._is_scrolling = True
            self._should_scroll = True
        else:
            if not self.bar.horizontal and not self.rotate:
                self.layout.width = self.bar.width
            elif self.scroll_fixed_width:
                self.length_type = bar.STATIC
                self.length = self._scroll_width
            else:
                self.length_type = bar.CALCULATED
            self._should_scroll = False

    def calculate_length(self):
        if not self.text:
            return 0
        if not self.bar.horizontal and not self.rotate:
            return self.layout.height + self.padding * 2
        else:
            return min(self.layout.width, self.bar.length) + self.padding * 2

    def can_draw(self):
        return self.layout is not None

    def rotate_drawer(self):
        if self.bar.horizontal or not self.rotate:
            return
        # Execute the base method when direction is default
        if self.direction == "default":
            _Widget.rotate_drawer(self)
        # Read bottom to top always with 'btt' direction
        elif self.direction == "btt":
            self.rotate_drawer_left()
        # Read top to bottom always with 'ttb' direction
        elif self.direction == "ttb":
            self.rotate_drawer_right()

    def draw(self):
        if not self.can_draw():
            return
        self.drawer.clear(self.background or self.bar.background)
        self.drawer.ctx.save()
        self.rotate_drawer()

        # If we're scrolling, we clip the context to the scroll width less the padding
        # Move the text layout position (and we only see the clipped portion)
        if self._should_scroll:
            height = self.bar.size if self.bar.horizontal or self.rotate else self.length
            self.drawer.ctx.rectangle(0, 0, self._scroll_width, height)
            self.drawer.ctx.clip()

        if not self.bar.horizontal and not self.rotate:
            x, y = 0, self.padding
        else:
            x = self.padding if self.length_type != bar.STATIC else 0
            y = (self.bar.size - self.layout.height) / 2 + 1

        self.layout.draw(x - self._scroll_offset, y)
        self.drawer.ctx.restore()

        self.draw_at_default_position()

        # We only want to scroll if:
        # - User has asked us to scroll and the scroll width is smaller than the layout (should_scroll=True)
        # - We are still scrolling (is_scrolling=True)
        # - We haven't already queued the next scroll (scroll_queued=False)
        if self._should_scroll and self._is_scrolling and not self._scroll_queued:
            self._scroll_queued = True
            if self._scroll_offset == 0:
                interval = self.scroll_delay
            else:
                interval = self.scroll_interval
            self._scroll_timer = self.timeout_add(interval, self.do_scroll)

    def do_scroll(self):
        # Allow the next scroll tick to be queued
        self._scroll_queued = False

        # If we're still scrolling, adjust the next offset
        if self._is_scrolling:
            self._scroll_offset += self.scroll_step

        # Check whether we need to stop scrolling when:
        # - we've scrolled all the text off the widget (scroll_clear = True)
        # - the final pixel is visible (scroll_clear = False)
        if (self.scroll_clear and self._scroll_offset > self.layout.width) or (
            not self.scroll_clear
            and (self.layout.width - self._scroll_offset) < (self._scroll_width)
        ):
            self._is_scrolling = False

        # We've reached the end of the scroll so what next?
        if not self._is_scrolling:
            if self.scroll_repeat:
                # Pause and restart scrolling
                self._scroll_timer = self.timeout_add(self.scroll_delay, self.reset_scroll)
            elif self.scroll_hide:
                # Clear the text
                self._scroll_timer = self.timeout_add(self.scroll_delay, self.hide_scroll)
            # If neither of these options then the text is no longer updated.

        self.draw()

    def reset_scroll(self):
        self._scroll_offset = 0
        self._is_scrolling = True
        self._scroll_queued = False
        if self._scroll_timer:
            self._scroll_timer.cancel()
        self.draw()

    def hide_scroll(self):
        self.update("")

    @expose_command()
    def set_font(
        self,
        font: str | None = None,
        fontsize: int = 0,
        fontshadow: ColorType = "",
        foreground: ColorType = "",
        markup: bool | None = None,
    ):
        """
        Change the text layout properties and redraw the widget.
        This method may also be used sync attributes from the current
        widget with the text layout.
        """
        if font is not None:
            self.font = font
        if fontsize != 0:
            self.fontsize = fontsize
        if fontshadow != "":
            self.fontshadow = fontshadow
        if foreground != "":
            self.foreground = foreground
        if markup is not None:
            self.markup = markup
        # Sync text layout properties
        if self.layout:
            self.layout.font_family = self.font
            self.layout.font_size = self.fontsize
            self.layout.font_shadow = self.fontshadow
            self.layout.colour = self.foreground
            self.layout.markup = self.markup
        self.bar.draw()

    @expose_command()
    def info(self):
        d = _Widget.info(self)
        d["text"] = self.formatted_text
        return d

    def update(self, text):
        """Update the widget text."""
        # Don't try to update text in dead layouts
        # This is mainly required for BackgroundPoll based widgets as the
        # polling function cannot be cancelled and so may be called after the widget
        # is finalised.
        if not self.can_draw():
            return

        if self.text == text:
            return
        if text is None:
            text = ""

        old_width = self.layout.width
        self.text = text

        # If our width hasn't changed, we just draw ourselves. Otherwise,
        # we draw the whole bar.
        if self.layout.width == old_width and (self.bar.horizontal or self.rotate):
            self.draw()
        else:
            self.bar.draw()


class InLoopPollText(_TextBox):
    """A common interface for polling some 'fast' information, munging it, and
    rendering the result in a text box. You probably want to use
    BackgroundPoll instead.

    ('fast' here means that this runs /in/ the event loop, so don't block! If
    you want to run something nontrivial, use BackgroundPoll.)"""

    defaults = [
        (
            "update_interval",
            600,
            "Update interval in seconds, if none, the widget updates only once.",
        ),
    ]  # type: list[tuple[str, Any, str]]

    def __init__(self, default_text="N/A", **config):
        _TextBox.__init__(self, default_text, **config)
        self.add_defaults(InLoopPollText.defaults)

    def timer_setup(self):
        update_interval = self.tick()
        # If self.update_interval is defined and .tick() returns None, re-call
        # after self.update_interval
        if update_interval is None and self.update_interval is not None:
            self.timeout_add(self.update_interval, self.timer_setup)
        # We can change the update interval by returning something from .tick()
        elif update_interval:
            self.timeout_add(update_interval, self.timer_setup)
        # If update_interval is False, we won't re-call

    def button_press(self, x, y, button):
        self.tick()
        _TextBox.button_press(self, x, y, button)

    def poll(self):
        return "N/A"

    def tick(self):
        text = self.poll()
        self.update(text)

    @expose_command()
    def force_update(self):
        """Immediately poll the widget. Existing timers are unaffected."""
        self.tick()


class BackgroundPoll(_TextBox):
    """A common interface for wrapping blocking events which when triggered
    will update a textbox.

    The poll/apoll methods are intended to wrap a blocking function which may
    take quite a while to return anything. Either method should return the
    string to update the widget text to. It may also return None to disable
    any further updates.

    If an `async def apoll()` is defined, that will be used to do the polling.

    For widgets that have not been ported to asyncio and define a `def poll()`
    method, their poll method will still be run in a thread as it is today.

    param: text - Initial text to display.
    """

    defaults = [
        (
            "update_interval",
            600,
            "Update interval in seconds, if none, the widget updates only once.",
        ),
    ]  # type: list[tuple[str, Any, str]]

    def __init__(self, text="N/A", **config):
        super().__init__(text, **config)
        self.add_defaults(BackgroundPoll.defaults)

    def timer_setup(self):
        create_task(self.do_tick())

    def poll(self):
        """An optional non-async-based method for polling. Will be run as an
        async future."""

    async def apoll(self):
        """An optional async-based method for polling."""

    async def do_tick(self, requeue=True):
        if type(self).apoll != BackgroundPoll.apoll:
            result = await self.apoll()
        elif type(self).poll != BackgroundPoll.poll:
            future = self.qtile.run_in_executor(self.poll)
            result = await future
        else:
            raise Exception(f"widget {self.name} has neither apoll() nor poll() overridden?")
        if result is not None:
            try:
                self.update(result)
            except Exception:
                logger.exception("Failed to reschedule timer for %s.", self.name)
            if requeue and self.update_interval is not None:
                await asyncio.sleep(self.update_interval)
                create_task(self.do_tick())
        else:
            logger.warning("%s's poll() returned None, not rescheduling", self.name)

    @expose_command()
    def force_update(self):
        """Immediately poll the widget. Existing timers are unaffected."""
        create_task(self.do_tick(requeue=False))


class PaddingMixin(configurable.Configurable):
    """Mixin that provides padding(_x|_y|)."""

    defaults = [
        ("padding", 3, "Padding inside the box"),
        ("padding_x", None, "X Padding. Overrides 'padding' if set"),
        ("padding_y", None, "Y Padding. Overrides 'padding' if set"),
    ]  # type: list[tuple[str, Any, str]]

    padding_x = configurable.ExtraFallback("padding_x", "padding")
    padding_y = configurable.ExtraFallback("padding_y", "padding")

    @property
    def padding_side(self):
        if self.bar.horizontal:
            return self.padding_x
        return self.padding_y

    @property
    def padding_top(self):
        if self.bar.horizontal:
            return self.padding_y
        return self.padding_x


class MarginMixin(configurable.Configurable):
    """Mixin that provides margin(_x|_y|)."""

    defaults = [
        ("margin", 3, "Margin inside the box"),
        ("margin_x", None, "X Margin. Overrides 'margin' if set"),
        ("margin_y", None, "Y Margin. Overrides 'margin' if set"),
    ]  # type: list[tuple[str, Any, str]]

    margin_x = configurable.ExtraFallback("margin_x", "margin")
    margin_y = configurable.ExtraFallback("margin_y", "margin")

    @property
    def margin_side(self):
        if self.bar.horizontal:
            return self.margin_x
        return self.margin_y

    @property
    def margin_top(self):
        if self.bar.horizontal:
            return self.margin_y
        return self.margin_x


class Mirror(_Widget):
    """
    A widget for showing the same widget content in more than one place, for
    instance, on bars across multiple screens.

    You don't need to use it directly; instead, just instantiate your widget
    once and hand it in to multiple bars. For instance::

        cpu = widget.CPUGraph()
        clock = widget.Clock()

        screens = [
            Screen(top=bar.Bar([widget.GroupBox(), cpu, clock])),
            Screen(top=bar.Bar([widget.GroupBox(), cpu, clock])),
        ]

    Widgets can be passed to more than one bar, so that there don't need to be
    any duplicates executing the same code all the time, and they'll always be
    visually identical.

    This works for all widgets that use `drawers` (and nothing else) to display
    their contents. Currently, this is all widgets except for `Systray`.
    """

    def __init__(self, reflection, **config):
        _Widget.__init__(self, reflection.length, **config)
        self.reflects = reflection
        self._length = 0
        self.length_type = self.reflects.length_type
        if self.length_type is bar.STATIC:
            self._length = self.reflects._length

    def _configure(self, qtile, bar):
        _Widget._configure(self, qtile, bar)
        self.reflects.add_mirror(self)
        # We need to fill the background once before `draw` is called so, if
        # there's no reflection, the mirror matches its parent bar.
        self.drawer.clear(self.background or self.bar.background)

    def calculate_length(self):
        return self.reflects.calculate_length()

    @property
    def length(self):
        if self.length_type != bar.STRETCH:
            return self.reflects.length
        return self._length

    @length.setter
    def length(self, value):
        self._length = value

    def draw(self):
        if self.length <= 0:
            return
        self.drawer.clear_rect()
        self.reflects.drawer.paint_to(self.drawer)
        self.draw_at_default_position()

    def button_press(self, x, y, button):
        self.reflects.button_press(x, y, button)

    def mouse_enter(self, x, y):
        self.reflects.mouse_enter(x, y)

    def mouse_leave(self, x, y):
        self.reflects.mouse_leave(x, y)

    def finalize(self):
        self.reflects.remove_mirror(self)
        _Widget.finalize(self)



from __future__ import annotations

from typing import TYPE_CHECKING

from libqtile.command.base import expose_command
from libqtile.layout.base import Layout

if TYPE_CHECKING:
    from collections.abc import Generator
    from typing import Any, Self

    from libqtile.backend.base import Window
    from libqtile.config import ScreenRect
    from libqtile.group import _Group


class _BspNode:
    def __init__(self, parent: _BspNode | None = None) -> None:
        self.parent = parent
        self.children: list[_BspNode] = []
        self.split_horizontal: bool = False
        self.split_ratio: float = 50
        self.client: Window | None = None
        self.x: int = 0
        self.y: int = 0
        self.w: int = 16
        self.h: int = 9

    def __iter__(self) -> Generator[_BspNode, None, None]:
        yield self
        for child in self.children:
            yield from child

    def clients(self) -> Generator[Window, None, None]:
        if self.client:
            yield self.client
        else:
            for child in self.children:
                yield from child.clients()

    def _shortest(self, length: int) -> tuple[_BspNode, int]:
        if len(self.children) == 0:
            return self, length

        child0, length0 = self.children[0]._shortest(length + 1)
        child1, length1 = self.children[1]._shortest(length + 1)

        if length1 < length0:
            return child1, length1
        return child0, length0

    def get_shortest(self) -> _BspNode:
        node, _ = self._shortest(0)
        return node

    def insert(self, client: Window, idx: int, ratio: float) -> _BspNode:
        if self.client is None:
            self.client = client
            return self
        self.children = [_BspNode(self), _BspNode(self)]
        self.children[1 - idx].client = self.client
        self.children[idx].client = client
        self.client = None
        self.split_horizontal = True if self.w > self.h * ratio else False
        return self.children[idx]

    def remove(self, child: _BspNode) -> _BspNode:
        keep = self.children[1 if child is self.children[0] else 0]
        self.children = keep.children
        for c in self.children:
            c.parent = self
        self.split_horizontal = keep.split_horizontal
        self.split_ratio = keep.split_ratio
        self.client = keep.client
        return self

    def distribute(self) -> tuple[int, int]:
        if len(self.children) == 0:
            return 1, 1
        h0, v0 = self.children[0].distribute()
        h1, v1 = self.children[1].distribute()
        if self.split_horizontal:
            h = h0 + h1
            v = max(v0, v1)
            self.split_ratio = 100 * h0 / h
        else:
            h = max(h0, h1)
            v = v0 + v1
            self.split_ratio = 100 * v0 / v
        return h, v

    def calc_geom(self, x: int, y: int, w: int, h: int) -> None:
        self.x = x
        self.y = y
        self.w = w
        self.h = h
        if len(self.children) > 1:
            if self.split_horizontal:
                w0 = int(self.split_ratio * w * 0.01 + 0.5)
                self.children[0].calc_geom(x, y, w0, h)
                self.children[1].calc_geom(x + w0, y, w - w0, h)
            else:
                h0 = int(self.split_ratio * h * 0.01 + 0.5)
                self.children[0].calc_geom(x, y, w, h0)
                self.children[1].calc_geom(x, y + h0, w, h - h0)


class Bsp(Layout):
    """This layout is inspired by bspwm, but it does not try to copy its
    features.

    The first client occupies the entire screen space.  When a new client
    is created, the selected space is partitioned in 2 and the new client
    occupies one of those subspaces, leaving the old client with the other.

    The partition can be either horizontal or vertical according to the
    dimensions of the current space: if its width/height ratio is above
    a pre-configured value, the subspaces are created side-by-side,
    otherwise, they are created on top of each other.  The partition
    direction can be freely toggled.  All subspaces can be resized and
    clients can be shuffled around.

    All clients are organized at the leaves of a full binary tree.

    An example key configuration is::

        Key([mod], "j", lazy.layout.down()),
        Key([mod], "k", lazy.layout.up()),
        Key([mod], "h", lazy.layout.left()),
        Key([mod], "l", lazy.layout.right()),
        Key([mod, "shift"], "j", lazy.layout.shuffle_down()),
        Key([mod, "shift"], "k", lazy.layout.shuffle_up()),
        Key([mod, "shift"], "h", lazy.layout.shuffle_left()),
        Key([mod, "shift"], "l", lazy.layout.shuffle_right()),
        Key([mod, "mod1"], "j", lazy.layout.flip_down()),
        Key([mod, "mod1"], "k", lazy.layout.flip_up()),
        Key([mod, "mod1"], "h", lazy.layout.flip_left()),
        Key([mod, "mod1"], "l", lazy.layout.flip_right()),
        Key([mod, "control"], "j", lazy.layout.grow_down()),
        Key([mod, "control"], "k", lazy.layout.grow_up()),
        Key([mod, "control"], "h", lazy.layout.grow_left()),
        Key([mod, "control"], "l", lazy.layout.grow_right()),
        Key([mod, "shift"], "n", lazy.layout.normalize()),
        Key([mod], "Return", lazy.layout.toggle_split()),
    """

    defaults = [
        ("border_focus", "#881111", "Border colour(s) for the focused window."),
        ("border_normal", "#220000", "Border colour(s) for un-focused windows."),
        ("border_width", 2, "Border width."),
        ("border_on_single", False, "Draw border when there is only one window."),
        (
            "margin_on_single",
            None,
            "Margin when there is only one window (int or list of ints [N E S W], 'None' to use 'margin' value).",
        ),
        ("margin", 0, "Margin of the layout (int or list of ints [N E S W])."),
        ("ratio", 1.6, "Width/height ratio that defines the partition direction."),
        ("grow_amount", 10, "Amount by which to grow a window/column."),
        ("lower_right", True, "New client occupies lower or right subspace."),
        ("fair", True, "New clients are inserted in the shortest branch."),
        (
            "wrap_clients",
            False,
            "Whether client list should be wrapped when using ``next`` and ``previous`` commands.",
        ),
    ]

    def __init__(self, **config):
        Layout.__init__(self, **config)
        self.add_defaults(Bsp.defaults)
        self.root = _BspNode()
        self.current = self.root
        if self.margin_on_single is None:
            self.margin_on_single = self.margin

    def swap(self, c1: Window, c2: Window) -> None:
        node_c1 = self.get_node(c1)
        node_c2 = self.get_node(c2)

        node_c1.client = c2
        node_c2.client = c1

        self.group.layout_all()

    def clone(self, group: _Group) -> Self:
        c = Layout.clone(self, group)
        c.root = _BspNode()
        c.current = c.root
        return c

    def get_windows(self):
        return list(self.root.clients())

    @expose_command()
    def info(self) -> dict[str, Any]:
        return dict(name=self.name, clients=[c.name for c in self.root.clients()])

    def get_node(self, client):
        for node in self.root:
            if client is node.client:
                return node

    def focus(self, client: Window) -> None:
        self.current = self.get_node(client)

    def add_client(self, client: Window) -> None:
        node = self.root.get_shortest() if self.fair else self.current
        self.current = node.insert(client, int(self.lower_right), self.ratio)

    def remove(self, client):
        node = self.get_node(client)
        if node:
            if node.parent:
                node = node.parent.remove(node)
                newclient = next(node.clients(), None)
                if newclient is None:
                    self.current = self.root
                else:
                    self.current = self.get_node(newclient)
                return newclient
            node.client = None
            self.current = self.root

    def configure(self, client: Window, screen_rect: ScreenRect) -> None:
        self.root.calc_geom(screen_rect.x, screen_rect.y, screen_rect.width, screen_rect.height)
        node = self.get_node(client)
        color = self.border_focus if client.has_focus else self.border_normal
        border = 0 if node is self.root and not self.border_on_single else self.border_width
        margin = self.margin_on_single if node is self.root else self.margin

        if node is not None:
            client.place(
                node.x,
                node.y,
                node.w - 2 * border,
                node.h - 2 * border,
                border,
                color,
                margin=margin,
            )
        client.unhide()

    @expose_command()
    def toggle_split(self):
        if self.current.parent:
            self.current.parent.split_horizontal = not self.current.parent.split_horizontal
        self.group.layout_all()

    def focus_first(self) -> Window | None:
        return next(self.root.clients(), None)

    def focus_last(self) -> Window | None:
        clients = list(self.root.clients())
        return clients[-1] if len(clients) else None

    def focus_next(self, client: Window, wrap: bool = False) -> Window | None:
        clients = list(self.root.clients())
        if client in clients:
            idx = clients.index(client)
            if not wrap and idx + 1 < len(clients):
                return clients[(idx + 1)]
            elif wrap:
                return clients[(idx + 1) % len(clients)]
        return None

    def focus_previous(self, client: Window, wrap: bool = False) -> Window | None:
        clients = list(self.root.clients())
        if client in clients:
            idx = clients.index(client)
            if not wrap and idx > 0:
                return clients[(idx - 1)]
            elif wrap:
                return clients[(idx - 1) % len(clients)]
        return None

    @expose_command()
    def next(self) -> None:
        client = self.focus_next(self.current.client, wrap=self.wrap_clients)
        if client:
            self.group.focus(client, True)

    @expose_command()
    def previous(self) -> None:
        client = self.focus_previous(self.current.client, wrap=self.wrap_clients)
        if client:
            self.group.focus(client, True)

    def find_left(self):
        child = self.current
        parent = child.parent
        while parent:
            if parent.split_horizontal and child is parent.children[1]:
                neighbor = parent.children[0]
                center = self.current.y + self.current.h * 0.5
                while neighbor.client is None:
                    if neighbor.split_horizontal or neighbor.children[1].y < center:
                        neighbor = neighbor.children[1]
                    else:
                        neighbor = neighbor.children[0]
                return neighbor
            child = parent
            parent = child.parent

    def find_right(self):
        child = self.current
        parent = child.parent
        while parent:
            if parent.split_horizontal and child is parent.children[0]:
                neighbor = parent.children[1]
                center = self.current.y + self.current.h * 0.5
                while neighbor.client is None:
                    if neighbor.split_horizontal or neighbor.children[1].y > center:
                        neighbor = neighbor.children[0]
                    else:
                        neighbor = neighbor.children[1]
                return neighbor
            child = parent
            parent = child.parent

    def find_up(self):
        child = self.current
        parent = child.parent
        while parent:
            if not parent.split_horizontal and child is parent.children[1]:
                neighbor = parent.children[0]
                center = self.current.x + self.current.w * 0.5
                while neighbor.client is None:
                    if not neighbor.split_horizontal or neighbor.children[1].x < center:
                        neighbor = neighbor.children[1]
                    else:
                        neighbor = neighbor.children[0]
                return neighbor
            child = parent
            parent = child.parent

    def find_down(self):
        child = self.current
        parent = child.parent
        while parent:
            if not parent.split_horizontal and child is parent.children[0]:
                neighbor = parent.children[1]
                center = self.current.x + self.current.w * 0.5
                while neighbor.client is None:
                    if not neighbor.split_horizontal or neighbor.children[1].x > center:
                        neighbor = neighbor.children[0]
                    else:
                        neighbor = neighbor.children[1]
                return neighbor
            child = parent
            parent = child.parent

    @expose_command()
    def left(self):
        node = self.find_left()
        if node:
            self.group.focus(node.client, True)

    @expose_command()
    def right(self):
        node = self.find_right()
        if node:
            self.group.focus(node.client, True)

    @expose_command()
    def up(self):
        node = self.find_up()
        if node:
            self.group.focus(node.client, True)

    @expose_command()
    def down(self):
        node = self.find_down()
        if node:
            self.group.focus(node.client, True)

    @expose_command()
    def shuffle_left(self):
        node = self.find_left()
        if node:
            node.client, self.current.client = self.current.client, node.client
            self.current = node
            self.group.layout_all()
        elif self.current is not self.root:
            node = self.current
            self.remove(node.client)
            newroot = _BspNode()
            newroot.split_horizontal = True
            newroot.children = [node, self.root]
            self.root.parent = newroot
            node.parent = newroot
            self.root = newroot
            self.current = node
            self.group.layout_all()

    @expose_command()
    def shuffle_right(self):
        node = self.find_right()
        if node:
            node.client, self.current.client = self.current.client, node.client
            self.current = node
            self.group.layout_all()
        elif self.current is not self.root:
            node = self.current
            self.remove(node.client)
            newroot = _BspNode()
            newroot.split_horizontal = True
            newroot.children = [self.root, node]
            self.root.parent = newroot
            node.parent = newroot
            self.root = newroot
            self.current = node
            self.group.layout_all()

    @expose_command()
    def shuffle_up(self):
        node = self.find_up()
        if node:
            node.client, self.current.client = self.current.client, node.client
            self.current = node
            self.group.layout_all()
        elif self.current is not self.root:
            node = self.current
            self.remove(node.client)
            newroot = _BspNode()
            newroot.split_horizontal = False
            newroot.children = [node, self.root]
            self.root.parent = newroot
            node.parent = newroot
            self.root = newroot
            self.current = node
            self.group.layout_all()

    @expose_command()
    def shuffle_down(self):
        node = self.find_down()
        if node:
            node.client, self.current.client = self.current.client, node.client
            self.current = node
            self.group.layout_all()
        elif self.current is not self.root:
            node = self.current
            self.remove(node.client)
            newroot = _BspNode()
            newroot.split_horizontal = False
            newroot.children = [self.root, node]
            self.root.parent = newroot
            node.parent = newroot
            self.root = newroot
            self.current = node
            self.group.layout_all()

    @expose_command()
    def grow_left(self):
        child = self.current
        parent = child.parent
        while parent:
            if parent.split_horizontal and child is parent.children[1]:
                parent.split_ratio = max(5, parent.split_ratio - self.grow_amount)
                self.group.layout_all()
                break
            child = parent
            parent = child.parent

    @expose_command()
    def grow_right(self):
        child = self.current
        parent = child.parent
        while parent:
            if parent.split_horizontal and child is parent.children[0]:
                parent.split_ratio = min(95, parent.split_ratio + self.grow_amount)
                self.group.layout_all()
                break
            child = parent
            parent = child.parent

    @expose_command()
    def grow_up(self):
        child = self.current
        parent = child.parent
        while parent:
            if not parent.split_horizontal and child is parent.children[1]:
                parent.split_ratio = max(5, parent.split_ratio - self.grow_amount)
                self.group.layout_all()
                break
            child = parent
            parent = child.parent

    @expose_command()
    def grow_down(self):
        child = self.current
        parent = child.parent
        while parent:
            if not parent.split_horizontal and child is parent.children[0]:
                parent.split_ratio = min(95, parent.split_ratio + self.grow_amount)
                self.group.layout_all()
                break
            child = parent
            parent = child.parent

    @expose_command()
    def flip_left(self):
        child = self.current
        parent = child.parent
        while parent:
            if parent.split_horizontal and child is parent.children[1]:
                parent.children = parent.children[::-1]
                self.group.layout_all()
                break
            child = parent
            parent = child.parent

    @expose_command()
    def flip_right(self):
        child = self.current
        parent = child.parent
        while parent:
            if parent.split_horizontal and child is parent.children[0]:
                parent.children = parent.children[::-1]
                self.group.layout_all()
                break
            child = parent
            parent = child.parent

    @expose_command()
    def flip_up(self):
        child = self.current
        parent = child.parent
        while parent:
            if not parent.split_horizontal and child is parent.children[1]:
                parent.children = parent.children[::-1]
                self.group.layout_all()
                break
            child = parent
            parent = child.parent

    @expose_command()
    def flip_down(self):
        child = self.current
        parent = child.parent
        while parent:
            if not parent.split_horizontal and child is parent.children[0]:
                parent.children = parent.children[::-1]
                self.group.layout_all()
                break
            child = parent
            parent = child.parent

    @expose_command()
    def normalize(self):
        distribute = True
        for node in self.root:
            if node.split_ratio != 50:
                node.split_ratio = 50
                distribute = False
        if distribute:
            self.root.distribute()
        self.group.layout_all()



"""
The interface to execute commands on the command graph
"""

from __future__ import annotations

import traceback
import types
import typing
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any, Literal, Union, get_args, get_origin

from libqtile import hook, ipc
from libqtile.command.base import CommandError, CommandException, CommandObject, SelectError
from libqtile.command.graph import CommandGraphCall, CommandGraphNode
from libqtile.log_utils import logger
from libqtile.utils import ColorsType, ColorType  # noqa: F401

if TYPE_CHECKING:
    from libqtile.command.graph import SelectorType

SUCCESS = 0
ERROR = 1
EXCEPTION = 2


# these two mask their aliases from elsewhere in the tree (i.e.
# libqtile.extension.base._Extension, and libqtile.layout.base.Layout
#
# since neither of these have constructors from a single string, we can't
# really lift them to a type. probably nobody actually passes us layouts here
# so it doesn't matter, but in the event that it does, we can probably fix it
# up with some eval() hackery whackery.
class _Extension:
    pass


class Layout:
    pass


def format_selectors(selectors: list[SelectorType]) -> str:
    """Build the path to the selected command graph node"""
    path_elements = []
    for name, selector in selectors:
        if selector is not None:
            path_elements.append(f"{name}[{selector}]")
        else:
            path_elements.append(name)
    return ".".join(path_elements)


class CommandInterface(metaclass=ABCMeta):
    """Defines an interface which can be used to evaluate a given call on a command graph.

    The implementations of this may use, for example, an IPC call to access the
    running qtile instance remotely or directly access the qtile instance from
    within the same process, or it may return lazily evaluated results.
    """

    @abstractmethod
    def execute(self, call: CommandGraphCall, args: tuple, kwargs: dict) -> Any:
        """Execute the given call, returning the result of the execution

        Perform the given command graph call, calling the function with the
        given arguments and keyword arguments.

        Parameters
        ----------
        call: CommandGraphCall
            The call on the command graph that is to be performed.
        args:
            The arguments to pass into the command graph call.
        kwargs:
            The keyword arguments to pass into the command graph call.
        """

    @abstractmethod
    def has_command(self, node: CommandGraphNode, command: str) -> bool:
        """Check if the given command exists

        Parameters
        ----------
        node: CommandGraphNode
            The node to check for commands
        command: str
            The name of the command to check for

        Returns
        -------
        bool
            True if the command is resolved on the given node
        """

    @abstractmethod
    def has_item(self, node: CommandGraphNode, object_type: str, item: str | int) -> bool:
        """Check if the given item exists

        Parameters
        ----------
        node: CommandGraphNode
            The node to check for items
        object_type: str
            The type of object to check for items.
        command: str
            The name of the item to check for

        Returns
        -------
        bool
            True if the item is resolved on the given node
        """


class QtileCommandInterface(CommandInterface):
    """Execute the commands via the in process running qtile instance"""

    def __init__(self, command_object: CommandObject):
        """A command object that directly resolves commands

        Parameters
        ----------
        command_object: CommandObject
            The command object to use for resolving the commands and items
            against.
        """
        self._command_object = command_object
        self.locked = False
        hook.subscribe.locked(self.lock)
        hook.subscribe.unlocked(self.unlock)

    def lock(self):
        self.locked = True

    def unlock(self):
        self.locked = False

    def execute(self, call: CommandGraphCall, args: tuple, kwargs: dict) -> Any:
        """Execute the given call, returning the result of the execution

        Perform the given command graph call, calling the function with the
        given arguments and keyword arguments.

        Parameters
        ----------
        call: CommandGraphCall
            The call on the command graph that is to be performed.
        args:
            The arguments to pass into the command graph call.
        kwargs:
            The keyword arguments to pass into the command graph call.
        """
        obj = self._command_object.select(call.selectors)
        cmd = None
        try:
            cmd = obj.command(call.name)
        except SelectError:
            pass

        if cmd is None:
            return "No such command."

        if self.locked and not getattr(cmd, "_allow_when_locked", False):
            return "Session is locked."

        logger.debug("Command: %s(%s, %s)", call.name, args, kwargs)
        return cmd(self._command_object, *args, **kwargs)

    def has_command(self, node: CommandGraphNode, command: str) -> bool:
        """Check if the given command exists

        Parameters
        ----------
        node: CommandGraphNode
            The node to check for commands
        command: str
            The name of the command to check for

        Returns
        -------
        bool
            True if the command is resolved on the given node
        """
        obj = self._command_object.select(node.selectors)
        cmd = obj.command(command)
        return cmd is not None

    def has_item(self, node: CommandGraphNode, object_type: str, item: str | int) -> bool:
        """Check if the given item exists

        Parameters
        ----------
        node: CommandGraphNode
            The node to check for items
        object_type: str
            The type of object to check for items.
        item: str
            The name or index of the item to check for

        Returns
        -------
        bool
            True if the item is resolved on the given node
        """
        try:
            self._command_object.select(node.selectors + [(object_type, item)])
        except SelectError:
            return False
        return True


class IPCCommandInterface(CommandInterface):
    """Execute the resolved commands using the IPC connection to a running qtile instance"""

    def __init__(self, ipc_client: ipc.Client):
        """Build a command object which resolves commands through IPC calls

        Parameters
        ----------
        ipc_client: ipc.Client
            The client that is to be used to resolve the calls.
        """
        self._client = ipc_client

    def execute(self, call: CommandGraphCall, args: tuple, kwargs: dict) -> Any:
        """Execute the given call, returning the result of the execution

        Executes the given command over the given IPC client.  Returns the
        result of the execution.

        Parameters
        ----------
        call: CommandGraphCall
            The call on the command graph that is to be performed.
        args:
            The arguments to pass into the command graph call.
        kwargs:
            The keyword arguments to pass into the command graph call.
        """
        status, result = self._client.send(
            (call.parent.selectors, call.name, args, kwargs, call.lifted)
        )
        if status == SUCCESS:
            return result
        if status == ERROR:
            raise CommandError(result)
        raise CommandException(result)

    def has_command(self, node: CommandGraphNode, command: str) -> bool:
        """Check if the given command exists

        Resolves the allowed commands over the IPC interface, and returns a
        boolean indicating of the given command is valid.

        Parameters
        ----------
        node: CommandGraphNode
            The node to check for commands
        command: str
            The name of the command to check for

        Returns
        -------
        bool
            True if the command is resolved on the given node
        """
        cmd_call = node.call("commands")
        commands = self.execute(cmd_call, (), {})
        return command in commands

    def has_item(self, node: CommandGraphNode, object_type: str, item: str | int) -> bool:
        """Check if the given item exists

        Resolves the available commands for the given command node of the given
        command type.  Performs the resolution of the items through the given
        IPC client.

        Parameters
        ----------
        node: CommandGraphNode
            The node to check for items
        object_type: str
            The type of object to check for items.
        command: str
            The name of the item to check for

        Returns
        -------
        bool
            True if the item is resolved on the given node
        """
        items_call = node.call("items")
        _, items = self.execute(items_call, (object_type,), {})
        return items is not None and item in items


def lift_args(cmd, args, kwargs):
    """
    Lift args lifts the arguments to the type annotations on cmd's parameters.
    """

    def lift_arg(typ, arg):
        # for stuff like int | None, allow either
        if get_origin(typ) in [types.UnionType, Union]:
            for t in get_args(typ):
                if t == types.NoneType:
                    # special case None? I don't know what this looks like
                    # coming over IPC
                    if arg == "":
                        return None
                    if arg is None:
                        return None
                    continue

                try:
                    return lift_arg(t, arg)
                except TypeError:
                    pass
            # uh oh, we couldn't lift it to anything
            raise TypeError(f"{arg} is not a {typ}")

        # for literals, check that it is one of the valid strings
        if get_origin(typ) is Literal:
            if arg not in get_args(typ):
                raise TypeError(f"{arg} is not one of {get_origin(typ)}")
            return arg

        if typ is bool:
            # >>> bool("False") is True
            # True
            # ... but we want it to be false :)
            if arg == "True" or arg is True:
                return True
            if arg == "False" or arg is False:
                return False
            raise TypeError(f"{arg} is not a bool")

        if typ is Any:
            # can't do any lifting if we don't know the type
            return arg

        if typ in [_Extension, Layout]:
            # these are "complex" objects that can't be created with a
            # single string argument. we generally don't expect people to
            # be passing these over the command line, so let's ignore then.
            return arg

        if get_origin(typ) in [list, dict]:
            # again, we do not want to be in the business of parsing
            # lists/dicts of types out of strings; just pass on whatever we
            # got
            return arg

        return typ(arg)

    converted_args = []
    converted_kwargs = dict()

    params = typing.get_type_hints(cmd, globalns=globals())

    non_return_annotated_args = filter(lambda k: k != "return", params.keys())
    for param, arg in zip(non_return_annotated_args, args):
        converted_args.append(lift_arg(params[param], arg))

    # if not all args were annotated, we need to keep them anyway. note
    # that mixing some annotated and not annotated args will not work well:
    # we will reorder args here and cause problems. this is solveable but
    # somewhat ugly, and we can avoid it by always annotating all
    # parameters.
    #
    # if we really want to fix this, we
    # inspect.signature(foo).parameters.keys() gives us the ordered
    # parameters to reason about.
    if len(converted_args) < len(args):
        converted_args.extend(args[len(converted_args) :])

    for k, v in kwargs.items():
        # if this kwarg has a type annotation, use it
        if k in params:
            converted_kwargs[k] = lift_arg(params[k], v)
        else:
            converted_kwargs[k] = v

    return tuple(converted_args), converted_kwargs


class IPCCommandServer:
    """Execute the object commands for the calls that are sent to it"""

    def __init__(self, qtile) -> None:
        """Wrapper around the ipc server for communitacing with the IPCCommandInterface

        sets up the IPC server such that it will receive and send messages to
        and from the IPCCommandInterface.
        """
        self.qtile = qtile

    def call(
        self,
        data: tuple[list[SelectorType], str, tuple, dict, bool],
    ) -> tuple[int, Any]:
        """Receive and parse the given data"""
        selectors, name, args, kwargs, lifted = data
        try:
            obj = self.qtile.select(selectors)
            cmd = obj.command(name)
        except SelectError as err:
            sel_string = format_selectors(selectors)
            return ERROR, f"No object {err.name} in path '{sel_string}'"
        if not cmd:
            return ERROR, "No such command"

        logger.debug("Command: %s(%s, %s)", name, args, kwargs)

        if lifted:
            args, kwargs = lift_args(cmd, args, kwargs)

        # Check if method is bound, if itis, insert magic self
        if not hasattr(cmd, "__self__"):
            args = (obj,) + args

        if self.qtile.locked and not getattr(cmd, "_allow_when_locked", False):
            return ERROR, f"{name} cannot be called when session is locked."

        try:
            return SUCCESS, cmd(*args, **kwargs)
        except CommandError as err:
            return ERROR, err.args[0]
        except Exception:
            return EXCEPTION, traceback.format_exc().strip().split("\n")[-1]



import libcst as cst
import libcst.matchers as m

from libqtile.scripts.migrations._base import (
    Check,
    MigrationTransformer,
    _QtileMigrator,
    add_migration,
)


class WlanArgsTransformer(MigrationTransformer):
    @m.call_if_inside(m.Call(func=m.Name("Wlan")) | m.Call(func=m.Attribute(attr=m.Name("Wlan"))))
    @m.leave(m.Arg(keyword=m.Name("ethernet_message")))
    def update_wlan_args(self, original_node, updated_node) -> cst.Arg:
        """Changes 'ethernet_message' to 'ethernet_message_format'."""
        self.lint(
            original_node,
            "The 'ethernet_message' argument is deprecated and should be replaced with 'ethernet_message_format'.",
        )
        return updated_node.with_changes(keyword=cst.Name("ethernet_message_format"))


class WlanArgs(_QtileMigrator):
    ID = "UpdateWlanArgs"

    SUMMARY = "Updates ``Wlan`` argument signature."

    HELP = """
    The ``Wlan`` widget previously accepted a ``ethernet_message`` keyword argument. This has
    been deprecated and should be replaced with a keyword argument named ``ethernet_message_format``.

    For example:

    .. code:: python

        widget.Wlan(ethernet_message="eth")

    should be changed to:

    .. code::

        widget.Wlan(ethernet_message_format="eth")

    """

    AFTER_VERSION = "0.31.0"

    TESTS = [
        Check(
            """
            from libqtile import bar, widget
            from libqtile.widget import Wlan

            bar.Bar(
                [
                    Wlan(ethernet_message="eth"),
                    widget.Wlan(ethernet_message="eth"),
                ],
                20,
            )
            """,
            """
            from libqtile import bar, widget
            from libqtile.widget import Wlan

            bar.Bar(
                [
                    Wlan(ethernet_message_format="eth"),
                    widget.Wlan(ethernet_message_format="eth"),
                ],
                20,
            )
            """,
        ),
    ]

    visitor = WlanArgsTransformer


add_migration(WlanArgs)



from libqtile.command.base import expose_command
from libqtile.widget import base


class QuickExit(base._TextBox):
    """
    A button to shut down Qtile. When clicked, a countdown starts. Clicking
    the button again stops the countdown and prevents Qtile from shutting down.
    """

    defaults = [
        ("default_text", "[ shutdown ]", "The text displayed on the button."),
        ("countdown_format", "[ {} seconds ]", "The text displayed when counting down."),
        ("timer_interval", 1, "The countdown interval."),
        ("countdown_start", 5, "The number to count down from."),
    ]

    def __init__(self, **config):
        base._TextBox.__init__(self, "", **config)
        self.add_defaults(QuickExit.defaults)

        self.is_counting = False
        self.text = self.default_text
        self.countdown = self.countdown_start

        self.add_callbacks({"Button1": self.trigger})

    def __reset(self):
        self.is_counting = False
        self.countdown = self.countdown_start
        self.text = self.default_text
        self.timer.cancel()

    def update(self):
        if not self.is_counting:
            return

        self.countdown -= 1
        self.text = self.countdown_format.format(self.countdown)
        self.timer = self.timeout_add(self.timer_interval, self.update)
        self.draw()

        if self.countdown == 0:
            self.qtile.stop()
            return

    @expose_command()
    def trigger(self):
        if not self.is_counting:
            self.is_counting = True
            self.update()
        else:
            self.__reset()
            self.draw()



import copy


class Configurable:
    global_defaults = {}  # type: dict

    def __init__(self, **config):
        self._variable_defaults = {}
        self._user_config = config

    def add_defaults(self, defaults):
        """Add defaults to this object, overwriting any which already exist"""
        # Since we can't check for immutability reliably, shallow copy the
        # value. If a mutable value were set and it were changed in one place
        # it would affect all other instances, since this is typically called
        # on __init__
        self._variable_defaults.update((d[0], copy.copy(d[1])) for d in defaults)

    def __getattr__(self, name):
        if name == "_variable_defaults":
            raise AttributeError
        found, value = self._find_default(name)
        if found:
            setattr(self, name, value)
            return value
        else:
            cname = self.__class__.__name__
            raise AttributeError(f"{cname} has no attribute: {name}")

    def _find_default(self, name):
        """Returns a tuple (found, value)"""
        defaults = self._variable_defaults.copy()
        defaults.update(self.global_defaults)
        defaults.update(self._user_config)
        if name in defaults:
            return (True, defaults[name])
        else:
            return (False, None)


class ExtraFallback:
    """Adds another layer of fallback to attributes

    Used to look up a different attribute name
    """

    def __init__(self, name, fallback):
        self.name = name
        self.hidden_attribute = "_" + name
        self.fallback = fallback

    def __get__(self, instance, owner=None):
        retval = getattr(instance, self.hidden_attribute, None)

        if retval is None:
            _found, retval = Configurable._find_default(instance, self.name)

        if retval is None:
            retval = getattr(instance, self.fallback, None)

        return retval

    def __set__(self, instance, value):
        """Set own value to a hidden attribute of the object"""
        setattr(instance, self.hidden_attribute, value)



import socket
from subprocess import DEVNULL, run

from libqtile.log_utils import logger
from libqtile.widget import base


class NetUP(base.BackgroundPoll):
    """
    A widget to display whether the network connection is up or down by probing a host via ping
    or tcp connection.

    By default ``host`` parameter is set to ``None``.
    """

    defaults = [
        ("host", None, "Host to probe."),
        ("method", "ping", "tcp or ping."),
        ("port", 443, "TCP port."),
        ("update_interval", 30, "Update interval in seconds."),
        ("display_fmt", "NET {0}", "Display format."),
        ("up_foreground", "FFFFFF", "Font color when host is up."),
        ("down_foreground", "FF0000", "Font color when host is down."),
        ("up_string", "up", "String to display when host is up."),
        ("down_string", "down", "String to display when host is down."),
    ]

    def __init__(self, **config):
        base.BackgroundPoll.__init__(self, **config)
        self.add_defaults(NetUP.defaults)

    def is_host_empty(self):
        if not self.host:
            logger.error("Host is not set")
            return False
        return True

    def validate_method(self):
        if self.method == "ping" or self.method == "tcp":
            return True
        logger.error("Method is invalid")
        return False

    def validate_port(self):
        if not isinstance(self.port, int):
            logger.error("Port is invalid")
            return False
        if self.port >= 1 and self.port <= 65535:
            return True
        else:
            logger.error("Port is invalid")
            return False

    def check_ping(self):
        process = run(["ping", "-c", "1", self.host], stdout=DEVNULL, stderr=DEVNULL)
        return process.returncode

    def check_tcp(self):
        sc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sc.settimeout(1)
        try:
            returncode = sc.connect_ex((self.host, self.port))
        except OSError:
            returncode = -1
        finally:
            sc.close()
        return returncode

    def is_up(self):
        if self.method == "ping":
            if self.check_ping() == 0:
                return True
            return False
        if self.method == "tcp":
            if self.check_tcp() == 0:
                return True
            return False

    def poll(self):
        if (
            not self.is_host_empty()
            or not self.validate_method()
            or (self.method == "tcp" and not self.validate_port())
        ):
            return "N/A"

        if self.is_up():
            self.layout.colour = self.up_foreground
            return self.display_fmt.format(self.up_string)
        self.layout.colour = self.down_foreground
        return self.display_fmt.format(self.down_string)



import psutil

from libqtile.widget import base

__all__ = ["Memory"]


class Memory(base.InLoopPollText):
    """Display memory/swap usage.

    The following fields are available in the `format` string:

    - ``MemUsed``: Memory in use.
    - ``MemTotal``: Total amount of memory.
    - ``MemFree``: Amount of memory free.
    - ``Available``: Amount of memory available.
    - ``NotAvailable``: Equal to ``MemTotal`` - ``MemAvailable``
    - ``MemPercent``: Memory in use as a percentage.
    - ``Buffers``: Buffer amount.
    - ``Active``: Active memory.
    - ``Inactive``: Inactive memory.
    - ``Shmem``: Shared memory.
    - ``SwapTotal``: Total amount of swap.
    - ``SwapFree``: Amount of swap free.
    - ``SwapUsed``: Amount of swap in use.
    - ``SwapPercent``: Swap in use as a percentage.
    - ``mm``: Measure unit for memory.
    - ``ms``: Measure unit for swap.

    Widget requirements: psutil_.

    .. _psutil: https://pypi.org/project/psutil/

    """

    defaults = [
        ("format", "{MemUsed: .0f}{mm}/{MemTotal: .0f}{mm}", "Formatting for field names."),
        ("update_interval", 1.0, "Update interval for the Memory"),
        ("measure_mem", "M", "Measurement for Memory (G, M, K, B)"),
        ("measure_swap", "M", "Measurement for Swap (G, M, K, B)"),
    ]

    measures = {"G": 1024 * 1024 * 1024, "M": 1024 * 1024, "K": 1024, "B": 1}

    def __init__(self, **config):
        super().__init__("", **config)
        self.add_defaults(Memory.defaults)
        self.calc_mem = self.measures[self.measure_mem]
        self.calc_swap = self.measures[self.measure_swap]

    def poll(self):
        mem = psutil.virtual_memory()
        swap = psutil.swap_memory()
        val = {}
        val["MemUsed"] = mem.used / self.calc_mem
        val["MemTotal"] = mem.total / self.calc_mem
        val["MemFree"] = mem.free / self.calc_mem
        val["Available"] = mem.available / self.calc_mem
        val["NotAvailable"] = (mem.total - mem.available) / self.calc_mem
        val["MemPercent"] = mem.percent
        val["Buffers"] = mem.buffers / self.calc_mem
        val["Active"] = mem.active / self.calc_mem
        val["Inactive"] = mem.inactive / self.calc_mem
        val["Shmem"] = mem.shared / self.calc_mem
        val["SwapTotal"] = swap.total / self.calc_swap
        val["SwapFree"] = swap.free / self.calc_swap
        val["SwapUsed"] = swap.used / self.calc_swap
        val["SwapPercent"] = swap.percent
        val["mm"] = self.measure_mem
        val["ms"] = self.measure_swap

        return self.format.format(**val)



import asyncio
import re
import subprocess

from libqtile import bar
from libqtile.command.base import expose_command
from libqtile.log_utils import logger
from libqtile.utils import acall_process, create_task
from libqtile.widget import base

__all__ = [
    "Volume",
]

re_vol = re.compile(r"(\d?\d?\d?)%")


class VolumeBase(base._TextBox):
    orientations = base.ORIENTATION_HORIZONTAL
    defaults = [
        ("padding", 3, "Padding left and right. Calculated if None."),
        ("theme_path", None, "Path of the icons"),
        (
            "emoji",
            False,
            "Use emoji to display volume states, only if ``theme_path`` is not set."
            "The specified font needs to contain the correct unicode characters.",
        ),
        (
            "emoji_list",
            ["\U0001f507", "\U0001f508", "\U0001f509", "\U0001f50a"],
            "List of emojis/font-symbols to display volume states, only if ``emoji`` is set."
            " List contains 4 symbols, from lowest volume to highest.",
        ),
        ("mute_foreground", None, "Foreground color for mute volume."),
        ("mute_format", "M", "Format to display when volume is muted."),
        ("unmute_format", "{volume}%", "Format of text to display when volume is not muted."),
        (
            "step",
            2,
            "Volume change for up an down commands in percentage."
            "Only used if ``volume_up_command`` and ``volume_down_command`` are not set.",
        ),
    ]

    def __init__(self, **config):
        base._TextBox.__init__(self, "", **config)
        self.add_defaults(VolumeBase.defaults)
        self.surfaces = {}
        self.volume = None
        self.is_mute = False

    def _configure(self, qtile, parent_bar):
        if self.theme_path:
            self.length_type = bar.STATIC
            self.length = 0
        base._TextBox._configure(self, qtile, parent_bar)
        self.unmute_foreground = self.foreground

    def _update_drawer(self):
        if self.mute_foreground is not None:
            self.layout.colour = self.mute_foreground if self.is_mute else self.unmute_foreground

        if self.theme_path:
            self.drawer.clear(self.background or self.bar.background)
            if self.volume <= 0 or self.is_mute:
                img_name = "audio-volume-muted"
            elif self.volume <= 30:
                img_name = "audio-volume-low"
            elif self.volume < 80:
                img_name = "audio-volume-medium"
            else:  # self.volume >= 80:
                img_name = "audio-volume-high"

            self.drawer.ctx.set_source(self.surfaces[img_name])
            self.drawer.ctx.paint()
        elif self.emoji:
            if len(self.emoji_list) < 4:
                self.emoji_list = ["\U0001f507", "\U0001f508", "\U0001f509", "\U0001f50a"]
                logger.warning(
                    "Emoji list given has less than 4 items. Falling back to default emojis."
                )

            if self.volume <= 0 or self.is_mute:
                self.text = self.emoji_list[0]
            elif self.volume <= 30:
                self.text = self.emoji_list[1]
            elif self.volume < 80:
                self.text = self.emoji_list[2]
            elif self.volume >= 80:
                self.text = self.emoji_list[3]
        else:
            self.text = (
                self.mute_format if self.is_mute or self.volume < 0 else self.unmute_format
            ).format(volume=self.volume)

    def setup_images(self):
        from libqtile import images

        names = (
            "audio-volume-high",
            "audio-volume-low",
            "audio-volume-medium",
            "audio-volume-muted",
        )
        d_images = images.Loader(self.theme_path)(*names)
        new_height = self.bar.size - 2
        for name, img in d_images.items():
            img.resize(height=new_height)
            if img.width > self.length:
                self.length = img.width + self.padding * 2
            self.surfaces[name] = img.pattern

    def draw(self):
        if self.theme_path:
            self.draw_at_default_position()
        else:
            base._TextBox.draw(self)


class Volume(VolumeBase):
    """Widget that display and change volume

    By default, this widget uses ``amixer`` to get and set the volume so users
    will need to make sure this is installed. Alternatively, users may set the
    relevant parameters for the widget to use a different application.

    If theme_path is set it draw widget as icons.
    """

    defaults = [
        ("cardid", None, "Card Id"),
        ("device", "default", "Device Name"),
        ("channel", "Master", "Channel"),
        ("update_interval", 0.2, "Update time in seconds."),
        ("mute_command", None, "Mute command"),
        ("volume_app", None, "App to control volume"),
        ("volume_up_command", None, "Volume up command"),
        ("volume_down_command", None, "Volume down command"),
        (
            "get_volume_command",
            None,
            "Command to get the current volume. "
            "The expected output should include 1-3 numbers and a ``%`` sign.",
        ),
        ("check_mute_command", None, "Command to check mute status"),
        (
            "check_mute_string",
            "[off]",
            "String expected from check_mute_command when volume is muted."
            "When the output of the command matches this string, the"
            "audio source is treated as muted.",
        ),
    ]

    def __init__(self, **config):
        VolumeBase.__init__(self, **config)
        self.add_defaults(Volume.defaults)

        self.add_callbacks(
            {
                "Button1": self.mute,
                "Button3": self.run_app,
                "Button4": self.increase_vol,
                "Button5": self.decrease_vol,
            }
        )

    def timer_setup(self):
        create_task(self.do_volume())
        if self.theme_path:
            self.setup_images()

    def create_amixer_command(self, *args) -> str:
        cmd = ["amixer"]

        if self.cardid is not None:
            cmd.extend(["-c", str(self.cardid)])

        if self.device is not None:
            cmd.extend(["-D", str(self.device)])

        cmd.extend([x for x in args])
        return subprocess.list2cmdline(cmd)

    def button_press(self, x, y, button):
        base._TextBox.button_press(self, x, y, button)
        self.draw()

    async def do_volume(self):
        vol, muted = await self.get_volume()
        if vol != self.volume or muted != self.is_mute:
            self.volume = vol
            self.is_mute = muted
            # Update the underlying canvas size before actually attempting
            # to figure out how big it is and draw it.
            self._update_drawer()
            self.bar.draw()
        await asyncio.sleep(self.update_interval)
        create_task(self.do_volume())

    async def get_volume(self):
        try:
            if self.get_volume_command is not None:
                get_volume_cmd = self.get_volume_command
            else:
                get_volume_cmd = self.create_amixer_command("sget", self.channel)

            mixer_out = await acall_process(get_volume_cmd, shell=True)
        except subprocess.CalledProcessError:
            return -1, False

        check_mute = mixer_out
        if self.check_mute_command:
            check_mute = await acall_process(self.check_mute_command, shell=True)

        muted = self.check_mute_string in check_mute

        volgroups = re_vol.search(mixer_out)
        if volgroups:
            return int(volgroups.groups()[0]), muted
        else:
            # this shouldn't happen
            return -1, muted

    @expose_command()
    def increase_vol(self):
        if self.volume_up_command is not None:
            volume_up_cmd = self.volume_up_command
        else:
            volume_up_cmd = self.create_amixer_command(
                "-q", "sset", self.channel, f"{self.step}%+"
            )

        subprocess.call(volume_up_cmd, shell=True)

    @expose_command()
    def decrease_vol(self):
        if self.volume_down_command is not None:
            volume_down_cmd = self.volume_down_command
        else:
            volume_down_cmd = self.create_amixer_command(
                "-q", "sset", self.channel, f"{self.step}%-"
            )

        subprocess.call(volume_down_cmd, shell=True)

    @expose_command()
    def mute(self):
        if self.mute_command is not None:
            mute_cmd = self.mute_command
        else:
            mute_cmd = self.create_amixer_command("-q", "sset", self.channel, "toggle")

        subprocess.call(mute_cmd, shell=True)

    @expose_command()
    def run_app(self):
        if self.volume_app is not None:
            subprocess.Popen(self.volume_app, shell=True)
